import os

from neuralpy.helpers.logs import red


CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))
TEMPLATE_FOLDER_PATH = os.path.join(CURRENT_DIR, 'new_project_template')

MIN_COOKIECUTTER_VER = '0.8.0'


def run():
    try:
        from cookiecutter.main import cookiecutter
    except ImportError:
        print(red("Install `cookiecutter` >= {} before run "
                  "this command.\n".format(MIN_COOKIECUTTER_VER)))
        print("Copy/paste/run command below:")
        print("pip install cookiecutter>{}".format(MIN_COOKIECUTTER_VER))
        return

    cookiecutter(TEMPLATE_FOLDER_PATH)



from numpy import reshape

from neuralpy.core.properties import NonNegativeIntProperty
from neuralpy.layers import StepLayer
from neuralpy.network.base import BaseNetwork
from neuralpy.network.learning import UnsupervisedLearning


__all__ = ('BaseStepAssociative',)


class BaseAssociative(UnsupervisedLearning, BaseNetwork):
    def __init__(self, connection, **options):
        if len(connection) != 2:
            raise ValueError("Connection must contains only input and "
                             "output layers")

        super(BaseAssociative, self).__init__(connection, **options)

    def setup_defaults(self):
        del self.use_bias
        self.use_bias = False
        super(BaseAssociative, self).setup_defaults()

    def train(self, input_train, epochs=None, epsilon=None):
        if epsilon is not None:
            raise AttributeError("You can't converge this algorithm. Use "
                                 "`epochs` parameter.")
        return super(BaseAssociative, self).train(input_train, epochs, epsilon)

    def predict(self, input_data):
        result = input_data
        for layer in self.layers:
            result = layer.output(result)
        return result


class BaseStepAssociative(BaseAssociative):
    """ Base class for associative algorithms which have 2 layers and first
    one is has step function as activation.
    """
    n_unconditioned = NonNegativeIntProperty(default=1, min_size=1)

    def __init__(self, connection, **options):
        super(BaseStepAssociative, self).__init__(connection, **options)

        input_layer = self.input_layer
        n_unconditioned = self.n_unconditioned

        if not isinstance(input_layer, StepLayer):
            raise ValueError("Input layer must be `StepLayer`")

        if input_layer.input_size <= n_unconditioned:
            raise ValueError(
                "Number of uncondition features must be less than total "
                "number of features in network. #feature = {} and "
                "#unconditioned = {}".format(
                    input_layer.input_size,
                    n_unconditioned
                )
            )

    def train_epoch(self, input_train, target_train):
        weight = self.input_layer.weight
        unconditioned = self.n_unconditioned
        predict = self.predict
        weight_delta = self.weight_delta

        for input_row in input_train:
            input_row = reshape(input_row, (1, input_row.size))
            layer_output = predict(input_row)
            weight[unconditioned:, :] += weight_delta(input_row, layer_output)



from six import with_metaclass
from numpy import dot
from numpy.random import randn

from neuralpy.core.config import ConfigMeta, BaseConfigurable
from neuralpy.core.properties import (IntProperty, NumberBoundProperty,
                                      ArrayProperty, ChoiceProperty)
from neuralpy.network.connections import ChainConnection
from neuralpy.network.utils import add_bias_column
from neuralpy.layers.utils import random_orthogonal, random_bounded


__all__ = ('BaseLayer',)


GAUSSIAN = 'gauss'
BOUNDED = 'bounded'
ORTHOGONAL = 'ortho'


class LayerMeta(ConfigMeta):
    def __new__(cls, clsname, bases, attrs):
        if 'activation_function' in attrs:
            # Override `activation_function` in `staticmethod` by default.
            # Python 2 translate any assigned function as class method
            # and try call it with with argument `self` which broke
            # logic. For this reason we try make it static.
            attrs['activation_function'] = staticmethod(
                attrs['activation_function']
            )
        return super(LayerMeta, cls).__new__(cls, clsname, bases, attrs)


class BaseLayer(with_metaclass(LayerMeta, ChainConnection, BaseConfigurable)):
    """ Base class for all layers.

    Parameters
    ----------
    {layer_params}
    """
    __layer_params = """input_size : int
        Layer input size.
    weight : 2D array-like or None
        Define your layer weights. `None` means that your weights will be
        generate randomly dependence on property `init_method`.
        `None` by default.
    init_method : {'gauss', 'bounded', 'ortho'}
        Weight initialization method.
        `gauss` will generate random weights dependence on Standard
        Normal Distribution.
        `bounded` generate uniform random weghts in initialized bounds.
        `ortho` generate random orthogonal matrix.
    random_weight_bound : tuple of two int
        Available only for `init_method` eqaul to `bounded`, defaults
        to `(0, 1)`.
    """
    shared_docs = {'layer_params': __layer_params}

    input_size = IntProperty()
    weight = ArrayProperty(default=None)
    random_weight_bound = NumberBoundProperty(default=(0, 1))
    init_method = ChoiceProperty(default=GAUSSIAN,
                                 choices=[GAUSSIAN, BOUNDED, ORTHOGONAL])

    def __init__(self, input_size, **options):
        super(BaseLayer, self).__init__()

        self.input_size = input_size
        self.use_bias = False

        # Default variables which will change after initialization
        self.relate_to_layer = None
        self.size = None

        # If you will set class method function variable, python understend
        # that this is new class method and will call it with `self`
        # first parameter.
        if hasattr(self.__class__, 'activation_function'):
            self.activation_function = self.__class__.activation_function

        # Initialize default options
        BaseConfigurable.__init__(self, **options)

    def relate_to(self, right_layer):
        self.relate_to_layer = right_layer

    def initialize(self, with_bias=False):
        self.use_bias = with_bias
        size = self.input_size + self.use_bias
        self.size = (size, self.relate_to_layer.input_size)
        self.weight = self._init_weight()

    # --------------- Weights manipulations --------------- #

    def _init_weight(self):
        if self.weight is not None:
            return self.weight

        init_method = self.init_method

        if init_method == GAUSSIAN:
            return randn(*self.size)

        elif init_method == BOUNDED:
            return random_bounded(self.size, *self.random_weight_bound)

        elif init_method == ORTHOGONAL:
            return random_orthogonal(self.size)

    @property
    def weight_without_bias(self):
        if self.use_bias:
            return self.weight[1:, :]
        return self.weight

    # --------------- Layer operations --------------- #

    def summator(self, input_value):
        return dot(input_value, self.weight)

    def output(self, input_value):
        input_data = self.preformat_input(input_value)
        summated = self.summator(input_data)
        return self.activation_function(summated)

    def preformat_input(self, input_data):
        if self.use_bias:
            input_data = add_bias_column(input_data)
        return input_data

    def __repr__(self):
        return '{name}({size})'.format(name=self.__class__.__name__,
                                       size=self.input_size)



import logging
import operator

from neuralpy.core.config import Configurable
from neuralpy.core.properties import BoolProperty


__all__ = ('Verbose',)


logging.basicConfig(level=logging.INFO, format="%(message)s")
logger = logging.getLogger(__file__)


def red(text):
    return "\033[91m{}\033[0m".format(text)


def green(text):
    return "\033[92m{}\033[0m".format(text)


def gray(text):
    return "\033[90m{}\033[0m".format(text)


def bold(text):
    return "\033[1m{}\033[0;0m".format(text)


def underline(text):
    return "\033[4m{}\033[0;0m".format(text)


class CustomLogger(logging.Logger):
    def _build_log(self, name, text, color=green):
        return "[{name}] {text}".format(name=color(name.upper()), text=text)

    def log(self, name, text, *args, **kwargs):
        message = self._build_log(name, text, color=green)
        self.info(message, *args, **kwargs)

    def gray_log(self, name, text, *args, **kwargs):
        message = self._build_log(name, text, color=gray)
        self.info(message, *args, **kwargs)

    def header(self, text, *args, **kwargs):
        message = "\n{text}\n".format(text=underline(bold(text)))
        self.info(message, *args, **kwargs)

    def simple(self, text):
        self.info(text)

    def empty(self):
        self.info("")

    def error(self, text, *args, **kwargs):
        message = self._build_log('ERROR', text, color=red)
        super(CustomLogger, self).error(message, *args, **kwargs)

    def warning(self, text, *args, **kwargs):
        message = self._build_log('WARN', text, color=red)
        super(CustomLogger, self).warning(message, *args, **kwargs)

    def data(self, text):
        text = text.strip()
        lines = list(map(operator.methodcaller('strip'), text.splitlines()))
        first_line = lines.pop(0)

        self.simple(bold(first_line))

        for line in lines:
            self.simple("  {} {}".format(green('*'), line))


logging.setLoggerClass(CustomLogger)


class VerboseProperty(BoolProperty):
    def __set__(self, instance, value):
        instance.logs.propagate = value
        return super(VerboseProperty, self).__set__(instance, value)


class Verbose(Configurable):
    verbose = VerboseProperty(default=True)

    def __init__(self, **options):
        logger_name = str(id(self))
        self.logs = logging.getLogger(logger_name)
        super(Verbose, self).__init__(**options)



from __future__ import division

from numpy import dot, asmatrix, reshape, where
from scipy.sparse import lil_matrix

from neuralpy.core.properties import BetweenZeroAndOneProperty
from .backpropagation import Backpropagation


__all__ = ('HessianDiagonal',)


class HessianDiagonal(Backpropagation):
    """ Hissian diagonal is a Hessian algorithm approximation which require
    only computation of hessian matrix diagonal elements and makes it
    invertion much easier and faster.

    Parameters
    ----------
    min_eigenvalue : float
        Setup min eigenvalue for Hessian diagonale matrix. After few
        iteration elements would be extremly small and matrix inverse
        produce huge number in hessian diagonal elements. This
        parameter control diagonal elements size. Defaults to ``1e-10``.
    {optimizations}
    {raw_predict_param}
    {full_params}

    Methods
    -------
    {supervised_train}
    {full_methods}

    Examples
    --------
    Simple example

    >>> import numpy as np
    >>> from neuralpy import algorithms
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> hdnet = algorithms.HessianDiagonal(
    ...     (2, 3, 1),
    ...     verbose=False
    ... )
    >>> hdnet.train(x_train, y_train)

    Diabets dataset example

    >>> import numpy as np
    >>> from sklearn.cross_validation import train_test_split
    >>> from sklearn import datasets, preprocessing
    >>> from neuralpy import algorithms, layers
    >>> from neuralpy.functions import rmsle
    >>>
    >>> np.random.seed(0)
    >>>
    >>> dataset = datasets.load_diabetes()
    >>> data, target = dataset.data, dataset.target
    >>>
    >>> input_scaler = preprocessing.StandardScaler()
    >>> target_scaler = preprocessing.StandardScaler()
    >>>
    >>> x_train, x_test, y_train, y_test = train_test_split(
    ...     input_scaler.fit_transform(data),
    ...     target_scaler.fit_transform(target),
    ...     train_size=0.8
    ... )
    >>>
    >>> nw = algorithms.HessianDiagonal(
    ...     connection=[
    ...         layers.SigmoidLayer(10),
    ...         layers.SigmoidLayer(20),
    ...         layers.OutputLayer(1)
    ...     ],
    ...     step=1.5,
    ...     use_raw_predict_at_error=False,
    ...     shuffle_data=False,
    ...     verbose=False,
    ...     min_eigenvalue=1e-10
    ... )
    >>> nw.train(x_train, y_train, epochs=10)
    >>> y_predict = nw.predict(x_test)
    >>>
    >>> error = rmsle(target_scaler.inverse_transform(y_test),
    ...               target_scaler.inverse_transform(y_predict).round())
    >>> error
    0.50315919814691346

    See Also
    --------
    :network:`Backpropagation` : Backpropagation algorithm.
    """
    min_eigenvalue = BetweenZeroAndOneProperty(default=1e-10)

    def get_weight_delta(self, output_train, target_train):
        weight_deltas = []
        gradients = self.gradients = []
        state_delta = self.delta = []

        update_first_order = self.error.deriv(output_train, target_train)
        min_eigenvalue = self.min_eigenvalue
        prev_weight = None
        update_second_order = None

        for i, layer in enumerate(reversed(self.train_layers), start=1):
            summated_data = self.summated_data[-i]
            current_layer_input = self.layer_outputs[-i]
            weight = layer.weight_without_bias.T
            weight_shape = layer.weight.shape

            activation_function_deriv = layer.activation_function.deriv
            deriv = activation_function_deriv(summated_data)
            second_deriv = activation_function_deriv.deriv(summated_data)

            if i == 1:
                # For last layer update
                delta = deriv ** 2 - update_first_order * second_deriv
            else:
                # For the hidden layers
                update_first_order = update_first_order.dot(prev_weight)
                delta = (
                    deriv ** 2 * update_second_order +
                    update_first_order * second_deriv
                )

            update_first_order *= deriv
            update_second_order = delta.dot(weight ** 2)

            # Compute gradient
            gradient = current_layer_input.T.dot(update_first_order).ravel()
            gradients.insert(0, reshape(gradient, weight_shape))

            # Compute hessian matrix
            weight_delta = asmatrix(dot(current_layer_input.T ** 2, delta))
            hessain_shape = (weight_delta.size, weight_delta.size)
            inverted_hessian = lil_matrix(hessain_shape)
            # Inverse for diagonal matrix is just reciprocal
            # every diagonal element
            full_gradients = weight_delta.ravel().T
            full_gradients = where(
                full_gradients < min_eigenvalue,
                min_eigenvalue,
                full_gradients
            )
            inverted_hessian.setdiag(1 / weight_delta.ravel().T)
            weight_delta = inverted_hessian.dot(gradient)

            weight_deltas.insert(0, reshape(-weight_delta, weight_shape))
            state_delta.insert(0, delta)
            prev_weight = weight

        return weight_deltas



from __future__ import division

from numpy import (zeros, argmin, argwhere, take, sum as np_sum,
                   any as np_any, abs as np_abs)
from numpy.linalg import norm

from neuralpy.core.properties import NonNegativeIntProperty
from neuralpy.network.base import BaseNetwork
from neuralpy.network.connections import FAKE_CONNECTION
from neuralpy.network.types import Clustering
from neuralpy.network.learning import UnsupervisedLearning


__all__ = ('RBFKMeans',)


class RBFKMeans(UnsupervisedLearning, Clustering, BaseNetwork):
    """ Radial basis function K-means for clustering.

    Parameters
    ----------
    n_clusters : int
        number of clusters in dataset.
    {show_epoch}
    {shuffle_data}
    {full_signals}
    {verbose}

    Attributes
    ----------
    centers : numpy array [n_clusters, n_futures]
        After training this property will contain coordinates
        to cluster centers.

    Methods
    -------
    {unsupervised_train_epsilon}
    {full_methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neuralpy.algorithms import RBFKMeans
    >>>
    >>> data = np.array([
    ...     [0.11, 0.20],
    ...     [0.25, 0.32],
    ...     [0.64, 0.60],
    ...     [0.12, 0.42],
    ...     [0.70, 0.73],
    ...     [0.30, 0.27],
    ...     [0.43, 0.81],
    ...     [0.44, 0.87],
    ...     [0.12, 0.92],
    ...     [0.56, 0.67],
    ...     [0.36, 0.35],
    ... ])
    >>> rbfk_net = RBFKMeans(n_clusters=2, verbose=False)
    >>> rbfk_net.train(data, epsilon=1e-5)
    >>> rbfk_net.centers
    array([[ 0.228     ,  0.312     ],
           [ 0.48166667,  0.76666667]])
    >>>
    >>> new_data = np.array([[0.1, 0.1], [0.9, 0.9]])
    >>> rbfk_net.predict(new_data)
    array([[ 0.],
           [ 1.]])
    """
    n_clusters = NonNegativeIntProperty(min_size=2)

    def __init__(self, **options):
        self.centers = None
        super(RBFKMeans, self).__init__(FAKE_CONNECTION, **options)

    def setup_defaults(self):
        del self.use_bias
        del self.error
        del self.step
        super(RBFKMeans, self).setup_defaults()

    def predict(self, input_data):
        centers = self.centers
        classes = zeros((input_data.shape[0], 1))

        for i, value in enumerate(input_data):
            classes[i] = argmin(norm(centers - value, axis=1))

        return classes

    def train_epoch(self, input_train, target_train):
        centers = self.centers
        old_centers = centers.copy()
        output_train = self.predict(input_train)

        for i, center in enumerate(centers):
            positions = argwhere(output_train[:, 0] == i)

            if not np_any(positions):
                continue

            class_data = take(input_train, positions, axis=0)
            centers[i, :] = (1 / len(class_data)) * np_sum(class_data, axis=0)

        return np_abs(old_centers - centers)

    def train(self, input_train, epsilon=1e-5):
        n_clusters = self.n_clusters

        if input_train.shape[0] <= n_clusters:
            raise ValueError("Count of clusters must be less than count of "
                             "input data.")

        self.centers = input_train[:n_clusters, :].copy()
        super(RBFKMeans, self).train(input_train, epsilon=epsilon)



__all__ = ('LEARING_RATE_UPDATE', 'WEIGHT_UPDATE', 'optimization_types')


# Available optimization types
LEARING_RATE_UPDATE = 1
WEIGHT_UPDATE = 2


optimization_types = {
    LEARING_RATE_UPDATE: "Learning rate update",
    WEIGHT_UPDATE: "Weight update",
}



from scipy.optimize import line_search

from neuralpy.core.properties import (NonNegativeNumberProperty,
                                      BetweenZeroAndOneProperty)
from neuralpy.algorithms.utils import (matrix_list_in_one_vector,
                                       vector_to_list_of_matrix)
from .base import SingleStep


__all__ = ('WolfeSearch',)


class WolfeSearch(SingleStep):
    """ Wolfe line search for the step selection.

    Parameters
    ----------
    maxstep : float
        Maximum step value. Defaults to ``50``.
    c1 : float
        Parameter for Armijo condition rule. Defaults to ``1e-4``.
    c2 : float
        Parameter for curvature condition rule. Defaults to ``0.9``.

    Attributes
    ----------
    {first_step}

    Warns
    -----
    {bp_depending}

    Examples
    --------
    >>> import numpy as np
    >>> from sklearn import datasets, metrics
    >>> from sklearn.cross_validation import StratifiedShuffleSplit
    >>> from neuralpy import algorithms, layers
    >>>
    >>> np.random.seed(0)
    >>>
    >>> X, y = datasets.make_classification(n_samples=100, n_features=10,
    ...                                     random_state=33)
    >>> shuffle_split = StratifiedShuffleSplit(y, 1, train_size=0.6,
    ...                                        random_state=33)
    >>>
    >>> train_index, test_index = next(shuffle_split.__iter__())
    >>> x_train, x_test = X[train_index], X[test_index]
    >>> y_train, y_test = y[train_index], y[test_index]
    >>>
    ... qnnet = algorithms.QuasiNewton(
    ...     connection=[
    ...         layers.SigmoidLayer(10, init_method='ortho'),
    ...         layers.SigmoidLayer(20, init_method='ortho'),
    ...         layers.OutputLayer(1)
    ...     ],
    ...     step=0.1,
    ...     use_raw_predict_at_error=False,
    ...     shuffle_data=True,
    ...     show_epoch=20,
    ...     verbose=False,
    ...
    ...     update_function='bfgs',
    ...     h0_scale=5,
    ...     gradient_tol=1e-5,
    ...     optimizations=[algorithms.WolfeSearch]
    ... )
    >>> qnnet.train(x_train, y_train, x_test, y_test, epochs=10)
    >>> result = qnnet.predict(x_test).round()
    >>>
    >>> roc_curve_score = metrics.roc_auc_score(result, y_test)
    >>> metrics.roc_auc_score(result, y_test)
    0.91666666666666674
    """

    maxstep = NonNegativeNumberProperty(default=50)
    c1 = BetweenZeroAndOneProperty(default=1e-4)
    c2 = BetweenZeroAndOneProperty(default=0.9)

    def set_weights(self, new_weights):
        for layer, new_weight in zip(self.train_layers, new_weights):
            layer.weight = new_weight.copy()

    def check_updates(self, new_step):
        weights = vector_to_list_of_matrix(
            new_step,
            (layer.size for layer in self.train_layers)
        )
        self.set_weights(weights)
        predicted_output = self.predict(self.input_train)
        return self.error(predicted_output, self.target_train)

    def get_gradient_by_weights(self, weights):
        weights = vector_to_list_of_matrix(
            weights,
            (layer.size for layer in self.train_layers)
        )
        self.set_weights(weights)
        gradient = self.get_gradient(self.output_train,
                                     self.target_train)
        return matrix_list_in_one_vector(gradient)

    def update_weights(self, weight_deltas):
        real_weights = [layer.weight for layer in self.train_layers]

        weights_vector = matrix_list_in_one_vector(real_weights)
        gradients_vetor = matrix_list_in_one_vector(self.gradients)

        res = line_search(self.check_updates,
                          self.get_gradient_by_weights,
                          xk=weights_vector,
                          pk=matrix_list_in_one_vector(weight_deltas),
                          gfk=gradients_vetor,
                          amax=self.maxstep,
                          c1=self.c1,
                          c2=self.c2)

        step = (res[0] if res[0] is not None else self.step)
        # SciPy some times ignore `amax` argument and return
        # bigger result
        self.step = min(self.maxstep, step)
        self.set_weights(real_weights)

        return super(WolfeSearch, self).update_weights(weight_deltas)



import numpy as np

from neuralpy.algorithms.basics.base import SimpleTwoLayerNetwork


__all__ = ('LMS',)


class LMS(SimpleTwoLayerNetwork):
    """ LMS Neural Network. Algorithm has several names, including the
    Widrow-Hoff or Delta rule. Algorithm similar to :network:`Perceptron`
    Neural Network, but has different idea behind learning process.

    Parameters
    ----------
    {full_params}

    Methods
    -------
    {supervised_train}
    {full_methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neuralpy import algorithms
    >>>
    >>> input_data = np.array([[1, 0], [2, 2], [3, 3], [0, 0]])
    >>> target_data = np.array([[1], [-1], [-1], [1]])
    >>>
    >>> lmsnet = algorithms.LMS((2, 1), step=0.5, verbose=False)
    >>> lmsnet.train(input_data, target_data, epochs=200)
    >>> lmsnet.predict(np.array([[4, 4], [-1, -1]]))
    array([[-1],
           [ 1]])

    See Also
    --------
    :network:`Perceptron` : Perceptron Neural Network.
    """
    def get_weight_delta(self, output_train, target_train):
        input_data = self.input_data

        minimized_input = input_data / np.linalg.norm(input_data) ** 2
        return self.step * np.dot(minimized_input.T,
                                  target_train - self.summated)



from abc import abstractmethod

from neuralpy.core.properties import BoolProperty
from neuralpy.layers.utils import generate_layers
from neuralpy.network.base import BaseNetwork


__all__ = ('FeedForwardNetwork',)


class FeedForwardNetwork(BaseNetwork):
    """ Base class for fedd forward neural network.
    """

    __raw_predict_param = """use_raw_predict_at_error : bool
        If the value is specified as a ``True``, it means that learning
        algorithm will use the results obtained by the network without
        passing through the output layer. Defaults to ``False``.
    """
    shared_docs = {'raw_predict_param': __raw_predict_param}

    use_raw_predict_at_error = BoolProperty(default=False)

    def __init__(self, connection, **options):
        islist_of_integers = (
            isinstance(connection, (list, tuple)) and
            isinstance(connection[0], int)
        )
        if islist_of_integers:
            connection = generate_layers(list(connection))

        super(FeedForwardNetwork, self).__init__(connection, **options)

    # ----------------- Active Neural Network State ---------------- #

    @abstractmethod
    def get_weight_delta(self, output_train, target_train):
        pass

    @abstractmethod
    def update_weights(self, weight_delta):
        pass

    def after_weight_update(self, input_train, target_train):
        pass

    def train_epoch(self, input_train, target_train):
        output_train = self.predict_for_error(input_train)
        self.output_train = output_train

        self.weight_delta = self.get_weight_delta(output_train, target_train)

        self.update_weights(self.weight_delta)
        self.after_weight_update(input_train, target_train)

        return self.error(output_train, target_train)

    def raw_predict(self, input_data):
        layer_outputs = self.layer_outputs = []
        summated_data = self.summated_data = []

        for layer in self.train_layers:
            input_data = layer.preformat_input(input_data)
            layer_outputs.append(input_data)

            summated = layer.summator(input_data)
            summated_data.append(summated)

            input_data = layer.activation_function(summated)

        return input_data

    def predict(self, input_data):
        raw_output = self.raw_predict(input_data)
        return self.output_layer.output(raw_output)

    def predict_for_error(self, input_data):
        if self.use_raw_predict_at_error:
            return self.raw_predict(input_data)
        return self.predict(input_data)



from neuralpy.core.properties import NonNegativeNumberProperty
from .base import WeightUpdateConfigurable


__all__ = ('WeightDecay',)


class WeightDecay(WeightUpdateConfigurable):
    """ Weight decay algorithm penalizes large weights and limits the
    freedom in network. The algorithm is able to solve one of the possible
    problems of network overfitting.

    Parameters
    ----------
    decay_rate : float
        Controls the effect of penalties on the update network weights.
        Defaults to ``0.1``.

    Warns
    -----
    {bp_depending}

    Examples
    --------
    >>> from neuralpy import algorithms
    >>>
    >>> bpnet = algorithms.Backpropagation(
    ...     (2, 4, 1),
    ...     step=0.1,
    ...     verbose=False,
    ...     optimizations=[algorithms.WeightDecay]
    ... )
    >>>

    See Also
    --------
    :network:`WeightElimination`
    """
    decay_rate = NonNegativeNumberProperty(default=0.1)

    def layer_weight_update(self, delta, layer_number):
        weight_update = super(WeightDecay, self).layer_weight_update(
            delta, layer_number
        )

        weight = self.train_layers[layer_number].weight
        step = self.layer_step(layer_number)

        return -step * self.decay_rate * weight + weight_update



from __future__ import division

from numpy import where, exp, log, cosh, tanh as np_tanh, clip

from neuralpy.functions import with_derivative


__all__ = ('sigmoid', 'linear', 'signum', 'tanh', 'rectifier', 'softplus',
           'softmax')


# -----------------------------------------------------#
#             Non-deferentiable functions              #
# -----------------------------------------------------#


def signum(x, upper_value=1, lower_value=0):
    return where(x > 0, upper_value, lower_value)


def rectifier(x):
    return where(x < 0, 0, x)


# -----------------------------------------------------#
#              Functions with derivatives              #
# -----------------------------------------------------#


def linear_deriv(x):
    return 1


@with_derivative(linear_deriv)
def linear(x):
    return x


def softmax_deriv(x, temp=1):
    softmax_result = softmax(x, temp)
    return softmax_result * (1 - softmax_result)


@with_derivative(softmax_deriv)
def softmax(x, temp=1):
    exp_input = exp(x / temp)
    exp_sum = exp_input.sum(axis=1)
    return exp_input / exp_sum.reshape((x.shape[0], 1))


# -----------------------------------------------------#
#        Functions with second-order derivatives       #
# -----------------------------------------------------#


def second_softplus_deriv(x):
    exp_for_x = exp(x)
    return exp_for_x / (1 + exp_for_x) ** 2


@with_derivative(second_softplus_deriv)
def softplus_deriv(x):
    exp_for_x = exp(x)
    return exp_for_x / (1 + exp_for_x)


@with_derivative(softplus_deriv)
def softplus(x):
    return log(1 + exp(x))


def second_sigmoid_deriv(x, alpha=1):
    exp_value = exp(alpha * x)
    return (
        -alpha ** 2 * (exp_value * (exp_value - 1))
    ) / (
        (exp_value + 1) ** 3
    )


@with_derivative(second_sigmoid_deriv)
def sigmoid_deriv(x, alpha=1):
    sigmoig_output = sigmoid(x, alpha=alpha)
    return alpha * sigmoig_output * (1 - sigmoig_output)


@with_derivative(sigmoid_deriv)
def sigmoid(x, alpha=1.):
    output = 1 / (1 + exp(-alpha * x))
    # Fix for infinite results
    return clip(output, a_min=0, a_max=1)


def tanh_second_deriv(x, alpha=1):
    return -2 * alpha ** 2 * tanh(x, alpha) * (1 / cosh(alpha * x)) ** 2


@with_derivative(tanh_second_deriv)
def tanh_deriv(x, alpha=1):
    return alpha - alpha * tanh(x, alpha) ** 2


@with_derivative(tanh_deriv)
def tanh(x, alpha=1):
    return np_tanh(alpha * x)



from operator import mul

from numpy import sqrt, inner

from neuralpy.core.properties import ChoiceProperty
from neuralpy.algorithms.utils import (matrix_list_in_one_vector,
                                       vector_to_list_of_matrix)
from .backpropagation import Backpropagation


__all__ = ('ConjugateGradient',)


def fletcher_reeves(gradient_old, gradient_new, weight_old_delta):
    return (
        inner(gradient_new, gradient_new) /
        inner(gradient_old, gradient_old)
    )


def polak_ribiere(gradient_old, gradient_new, weight_old_delta):
    return (
        inner(gradient_new, gradient_new - gradient_old) /
        inner(gradient_old, gradient_old)
    )


def hentenes_stiefel(gradient_old, gradient_new, weight_old_delta):
    gradient_delta = gradient_new - gradient_old
    return (
        inner(gradient_delta, gradient_new) /
        inner(weight_old_delta, gradient_delta)
    )


def conjugate_descent(gradient_old, gradient_new, weight_old_delta):
    # Note: `sqrt(dot(a.T, a))` works faster than `linalg.norm(a)`
    return (
        -sqrt(inner(gradient_new, gradient_new)) /
        inner(weight_old_delta, gradient_old)
    )


def liu_storey(gradient_old, gradient_new, weight_old_delta):
    return (
        inner(gradient_new, gradient_new - gradient_old) /
        inner(weight_old_delta, gradient_old)
    )


def dai_yuan(gradient_old, gradient_new, weight_old_delta):
    return (
        inner(gradient_new, gradient_new) /
        inner(gradient_new - gradient_old, weight_old_delta)
    )


class ConjugateGradient(Backpropagation):
    """ Conjugate Gradient algorithm.

    Parameters
    ----------
    update_function : {{'fletcher_reeves', 'polak_ribiere',\
    'hentenes_stiefel', 'conjugate_descent', 'liu_storey', 'dai_yuan'}}
        Update function. Defaults to ``fletcher_reeves``.
    {optimizations}
    {raw_predict_param}
    {full_params}

    Methods
    -------
    {supervised_train}
    {full_methods}

    Examples
    --------
    >>> import numpy as np
    >>> np.random.seed(0)
    >>>
    >>> from sklearn import datasets, preprocessing
    >>> from sklearn.cross_validation import train_test_split
    >>> from neuralpy import algorithms, layers
    >>> from neuralpy.functions import rmsle
    >>>
    >>> dataset = datasets.load_boston()
    >>> data, target = dataset.data, dataset.target
    >>>
    >>> data_scaler = preprocessing.MinMaxScaler()
    >>> target_scaler = preprocessing.MinMaxScaler()
    >>>
    >>> x_train, x_test, y_train, y_test = train_test_split(
    ...     data_scaler.fit_transform(data),
    ...     target_scaler.fit_transform(target),
    ...     train_size=0.85
    ... )
    >>>
    >>> cgnet = algorithms.ConjugateGradient(
    ...     connection=[
    ...         layers.SigmoidLayer(13),
    ...         layers.SigmoidLayer(50),
    ...         layers.OutputLayer(1),
    ...     ],
    ...     search_method='golden',
    ...     update_function='fletcher_reeves',
    ...     optimizations=[algorithms.LinearSearch],
    ...     verbose=False
    ... )
    >>>
    >>> cgnet.train(x_train, y_train, epochs=100)
    >>> y_predict = cgnet.predict(x_test)
    >>>
    >>> real = target_scaler.inverse_transform(y_test)
    >>> predicted = target_scaler.inverse_transform(y_predict)
    >>>
    >>> error = rmsle(real, predicted.round(1))
    >>> error
    0.20752676697596578

    See Also
    --------
    :network:`Backpropagation`: Backpropagation algorithm.
    :network:`LinearSearch`: Linear Search important algorithm for step \
    selection in Conjugate Gradient algorithm.
    """
    update_function = ChoiceProperty(
        default='fletcher_reeves',
        choices={
            'fletcher_reeves': fletcher_reeves,
            'polak_ribiere': polak_ribiere,
            'hentenes_stiefel': hentenes_stiefel,
            'conjugate_descent': conjugate_descent,
            'liu_storey': liu_storey,
            'dai_yuan': dai_yuan,
        }
    )

    def init_layers(self):
        super(ConjugateGradient, self).init_layers()
        self.n_weights = sum(mul(*layer.size) for layer in self.train_layers)

    def get_weight_delta(self, output_train, target_train):
        gradients = super(ConjugateGradient, self).get_gradient(output_train,
                                                                target_train)
        epoch = self.epoch
        gradient = matrix_list_in_one_vector(gradients)
        weight_delta = -gradient

        if epoch > 1 and epoch % self.n_weights == 0:
            # Must reset after every N iteration, because algoritm
            # lose conjugacy.
            self.logs.info("TRAIN", "Reset conjugate gradient vector")
            del self.prev_gradient

        if hasattr(self, 'prev_gradient'):
            gradient_old = self.prev_gradient
            weight_delta_old = self.prev_weight_delta
            beta = self.update_function(gradient_old, gradient,
                                        weight_delta_old)

            weight_delta += beta * weight_delta_old

        weight_deltas = vector_to_list_of_matrix(
            weight_delta,
            (layer.size for layer in self.train_layers)
        )

        self.prev_weight_delta = weight_delta.copy()
        self.prev_gradient = gradient.copy()

        return weight_deltas



import copy

from neuralpy.core.properties import BetweenZeroAndOneProperty
from .backpropagation import Backpropagation


__all__ = ('Momentum',)


class Momentum(Backpropagation):
    """ Momentum algorithm for :network:`Backpropagation` optimization.

    Parameters
    ----------
    momentum : float
        Control previous gradient ratio. Defaults to ``0.9``.
    {optimizations}
    {raw_predict_param}
    {full_params}

    Methods
    -------
    {supervised_train}
    {full_methods}

    Examples
    --------
    Simple example

    >>> import numpy as np
    >>> from neuralpy import algorithms
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> mnet = algorithms.Momentum(
    ...     (2, 3, 1),
    ...     verbose=False
    ... )
    >>> mnet.train(x_train, y_train)

    See Also
    --------
    :network:`Backpropagation` : Backpropagation algorithm.
    """
    momentum = BetweenZeroAndOneProperty(default=0.9)

    def layer_weight_update(self, delta, layer_number):
        update = super(Momentum, self).layer_weight_update(delta,
                                                           layer_number)
        if not hasattr(self, 'prev_gradients'):
            return update
        return -self.momentum * self.prev_gradients[layer_number] + update

    def update_weights(self, weight_deltas):
        super(Momentum, self).update_weights(weight_deltas)
        self.prev_gradients = copy.copy(self.gradients)



from __future__ import division

from numpy import dot, zeros, ones, inf, logical_and, sort, unique

from neuralpy.core.properties import (BetweenZeroAndOneProperty,
                                      NonNegativeIntProperty)
from neuralpy.network.base import BaseNetwork
from neuralpy.network.types import Clustering
from neuralpy.network.connections import FAKE_CONNECTION


__all__ = ('ART1',)


class ART1(Clustering, BaseNetwork):
    """ Adaptive Resonance Theory (ART1) Network for binary
    data clustering.

    Notes
    -----
    * Weights are not random, so the result will be always reproduceble.

    Parameters
    ----------
    rho : float
        Control reset action in trainig process. Value must be
        between ``0`` and ``1``, defaults to ``0.5``.
    n_clusters : int
        Number of clusters, defaults to ``2``. Min value is also ``2``.
    {full_params}

    Methods
    -------
    train(input_data):
        Trains network until it has clustered all samples
    {predict}
    {plot_errors}
    {last_error}

    Examples
    --------
    >>> import numpy as np
    >>> from neuralpy import algorithms
    >>>
    >>> data = np.array([
    ...     [0, 1, 0],
    ...     [1, 0, 0],
    ...     [1, 1, 0],
    ... ])
    >>>>
    >>> artnet = algorithms.ART1(
    ...     step=2,
    ...     rho=0.7,
    ...     n_clusters=2,
    ...     verbose=False
    ... )
    >>> artnet.predict(data)
    array([ 0.,  1.,  1.])
    """
    rho = BetweenZeroAndOneProperty(default=0.5)
    n_clusters = NonNegativeIntProperty(default=2, min_size=2)

    def __init__(self, **options):
        super(ART1, self).__init__(FAKE_CONNECTION, **options)

    def train(self, input_data):
        if input_data.ndim != 2:
            raise ValueError("Input value must be 2 dimentional, got "
                             "{0}".format(input_data.ndim))

        data_size = input_data.shape[1]
        n_clusters = self.n_clusters
        step = self.step
        rho = self.rho

        if list(sort(unique(input_data))) != [0, 1]:
            raise ValueError("ART1 Network works only with binary matrix, "
                             "all matix must contains only 0 and 1")

        if not hasattr(self, 'weight_21'):
            self.weight_21 = ones((data_size, n_clusters))

        if not hasattr(self, 'weight_12'):
            self.weight_12 = step / (step + n_clusters - 1) * self.weight_21.T

        weight_21 = self.weight_21
        weight_12 = self.weight_12

        if data_size != weight_21.shape[0]:
            raise ValueError(
                "Data dimention is invalid. Get {} columns data set. "
                "Must be - {} columns".format(
                    data_size, weight_21.shape[0]
                )
            )

        classes = zeros(input_data.shape[0])

        # Train network
        for i, p in enumerate(input_data):
            disabled_neurons = []
            reseted_values = []
            reset = True

            while reset:
                output1 = p
                input2 = dot(weight_12, output1.T)

                output2 = zeros(input2.size)
                input2[disabled_neurons] = -inf
                winner_index = input2.argmax()
                output2[winner_index] = 1

                expectation = dot(weight_21, output2)
                output1 = logical_and(p, expectation).astype(int)

                reset_value = dot(output1.T, output1) / dot(p.T, p)
                reset = reset_value < rho

                if reset:
                    disabled_neurons.append(winner_index)
                    reseted_values.append((reset_value, winner_index))

                if len(disabled_neurons) >= n_clusters:
                    # Got this case only if we test all possible clusters
                    reset = False
                    winner_index = None

                if not reset:
                    if winner_index is not None:
                        weight_12[winner_index, :] = (step * output1) / (
                            step + dot(output1.T, output1) - 1
                        )
                        weight_21[:, winner_index] = output1
                    else:
                        # Get result with the best `rho`
                        winner_index = max(reseted_values)[1]

                    classes[i] = winner_index

        return classes

    def predict(self, input_data):
        return self.train(input_data)

    def train_epoch(self, input_data, target_data):
        pass



from neuralpy.core.config import Configurable
from neuralpy.algorithms.backprop import WEIGHT_UPDATE


__all__ = ('WeightUpdateConfigurable',)


class WeightUpdateConfigurable(Configurable):
    optimization_type = WEIGHT_UPDATE



from numpy import dot, abs as np_abs
from numpy.random import randn

from neuralpy.core.properties import NonNegativeIntProperty, ArrayProperty
from neuralpy.network.base import BaseNetwork
from neuralpy.network.learning import UnsupervisedLearning
from neuralpy.network.connections import FAKE_CONNECTION


__all__ = ('Oja',)


class Oja(UnsupervisedLearning, BaseNetwork):
    """ Oja unsupervised algorithm which minimize feature space.

    Notes
    -----
    * In practice use step as very small value. For example ``1e-7``.

    Parameters
    ----------
    minimized_data_size : int
        Expected number of features after minimization, defaults to ``1``
    weights : array-like or ``None``
        Predefine default weights which controll your data in two sides.
        If weights are, ``None`` before train algorithms generate random
        weights. Defaults to ``None``.
    {step}
    {show_epoch}
    {verbose}
    {full_signals}

    Methods
    -------
    train(input_train, epsilon=1e-5):
        Train network. As result it returns compresed data.
    reconstruct(input_data):
        Reconstruct your minimized data.
    {fit}
    {last_error}
    {plot_errors}

    Raises
    ------
    ValueError
        * Try reconstruct without training.
        * Invalid number of input data features for ``train`` and \
        ``reconstruct`` methods.

    Examples
    --------
    >>> import numpy as np
    >>> from neuralpy import algorithms
    >>>
    >>> data = np.array([[2, 2], [1, 1], [4, 4], [5, 5]])
    >>>
    >>> ojanet = algorithms.Oja(
    ...     minimized_data_size=1,
    ...     step=0.01,
    ...     verbose=False
    ... )
    >>>
    >>> minimized = ojanet.train(data, epsilon=1e-5)
    >>> minimized
    array([[-2.82843122],
           [-1.41421561],
           [-5.65686243],
           [-7.07107804]])
    >>> ojanet.reconstruct(minimized)
    array([[ 2.00000046,  2.00000046],
           [ 1.00000023,  1.00000023],
           [ 4.00000093,  4.00000093],
           [ 5.00000116,  5.00000116]])
    """
    minimized_data_size = NonNegativeIntProperty(min_size=1)
    weights = ArrayProperty()

    def __init__(self, **options):
        super(Oja, self).__init__(FAKE_CONNECTION, **options)

    def setup_defaults(self):
        del self.use_bias
        del self.error
        del self.shuffle_data
        super(Oja, self).setup_defaults()

    def train_epoch(self, input_data, target_train):
        weights = self.weights

        self.minimized = minimized = dot(input_data, weights)
        reconstruct = dot(minimized, weights.T)
        error = input_data - reconstruct

        weights += self.step * dot(error.T, minimized)

        return np_abs(error) / (input_data.shape[0] * input_data.shape[1])

    def train(self, input_data, epsilon=1e-5):
        self.minimized = None
        n_input_features = input_data.shape[1]

        if self.weights is None:
            self.weights = randn(n_input_features, self.minimized_data_size)

        if n_input_features != self.weights.shape[0]:
            raise ValueError(
                "Invalid number of features. Expected {}, got {}".format(
                    self.weights.shape[0], n_input_features
                )
            )

        super(Oja, self).train(input_data, epsilon=epsilon)

        return self.minimized

    def reconstruct(self, input_data):
        if self.weights is None:
            raise ValueError("Train network before reconstruct data.")

        if input_data.shape[1] != self.minimized_data_size:
            raise ValueError(
                "Invalid input data feature space, expected "
                "{}, got {}.".format(
                    input_data.shape[1], self.minimized_data_size
                )
            )

        return dot(input_data, self.weights.T)

    def predict(self, input_data):
        raise AttributeError("Can't predict value for Oja network.")



import inspect
import importlib

import numpy as np


__all__ = ('preformat_value', 'import_class')


def preformat_value(value):
    if inspect.isfunction(value) or inspect.isclass(value):
        return value.__name__

    elif isinstance(value, (list, tuple, set)):
        return [preformat_value(v) for v in value]

    elif isinstance(value, (np.ndarray, np.matrix)):
        return value.shape

    return value


def import_class(path_to_class):
    module_name, classname = path_to_class.rsplit('.', 1)
    module = importlib.import_module(module_name)
    return getattr(module, classname)



from abc import abstractmethod

from neuralpy.helpers.logs import Verbose
from .config import ConfigurableWithABC


__all__ = ('BaseSkeleton',)


class BaseSkeleton(ConfigurableWithABC, Verbose):
    """ Base class for all algorithms and networks.
    """
    @abstractmethod
    def train(self, input_data, target_data):
        pass

    @abstractmethod
    def predict(self, input_data):
        pass

    def fit(self, X, y, *args, **kwargs):
        self.train(X, y, *args, **kwargs)
        return self



from math import log, ceil

from numpy import zeros, fill_diagonal, random, multiply

from neuralpy.core.properties import ChoiceProperty, NonNegativeIntProperty
from neuralpy.functions import signum
from .utils import bin2sign
from .base import DiscreteMemory


__all__ = ('DiscreteHopfieldNetwork',)


class DiscreteHopfieldNetwork(DiscreteMemory):
    """ Discrete Hopfield Network. Memory algorithm which works only with
    binary vectors.

    Notes
    -----
    * {discrete_data_note}

    Examples
    --------
    >>> import numpy as np
    >>> from neuralpy import algorithms
    >>>
    >>> def draw_bin_image(image_matrix):
    ...     for row in image_matrix.tolist():
    ...         print('| ' + ' '.join(' *'[val] for val in row))
    ...
    >>> zero = np.matrix([
    ...     0, 1, 1, 1, 0,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     0, 1, 1, 1, 0
    ... ])
    >>>
    >>> one = np.matrix([
    ...     0, 1, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0
    ... ])
    >>>
    >>> two = np.matrix([
    ...     1, 1, 1, 0, 0,
    ...     0, 0, 0, 1, 0,
    ...     0, 0, 0, 1, 0,
    ...     0, 1, 1, 0, 0,
    ...     1, 0, 0, 0, 0,
    ...     1, 1, 1, 1, 1,
    ... ])
    >>>
    >>> half_zero = np.matrix([
    ...     0, 1, 1, 1, 0,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     0, 0, 0, 0, 0,
    ...     0, 0, 0, 0, 0,
    ...     0, 0, 0, 0, 0,
    ... ])
    >>>
    >>> draw_bin_image(zero.reshape((6, 5)))
    |   * * *
    | *       *
    | *       *
    | *       *
    | *       *
    |   * * *
    >>> draw_bin_image(half_zero.reshape((6, 5)))
    |   * * *
    | *       *
    | *       *
    |
    |
    |
    >>> data = np.concatenate([zero, one, two], axis=0)
    >>>
    >>> dhnet = algorithms.DiscreteHopfieldNetwork()
    >>> dhnet.train(data)
    >>>
    >>> result = dhnet.predict(half_zero)
    >>> draw_bin_image(result.reshape((6, 5)))
    |   * * *
    | *       *
    | *       *
    | *       *
    | *       *
    |   * * *
    """
    mode = ChoiceProperty(default='full', choices=['random', 'full'])
    n_nodes = NonNegativeIntProperty(default=100)

    def __init__(self, **options):
        super(DiscreteHopfieldNetwork, self).__init__(**options)
        self.n_remembered_data = 0
        self.weight = None

        if 'n_nodes' in options and self.mode != 'random':
            self.logs.warning("You can use `n_nodes` property only in "
                              "`random` mode.")

    def train(self, input_data):
        self.discrete_validation(input_data)

        input_data = bin2sign(input_data)

        nrows, n_features = input_data.shape
        nrows_after_update = self.n_remembered_data + nrows
        memory_limit = ceil(n_features / (2 * log(n_features)))

        if nrows_after_update > memory_limit:
            raise ValueError("You can't memorize more than {0} "
                             "samples".format(memory_limit))

        weight_shape = (n_features, n_features)

        if self.weight is None:
            self.weight = zeros(weight_shape)

        if self.weight.shape != weight_shape:
            raise ValueError("Invalid input shapes. Number of input "
                             "features must be equal to {} and output "
                             "features - {}".format(*weight_shape))

        self.weight = input_data.T.dot(input_data)
        fill_diagonal(self.weight, zeros(len(self.weight)))
        self.n_remembered_data = nrows_after_update

    def predict(self, input_data):
        input_data = bin2sign(input_data)

        if self.mode == 'random':
            # Valid number of features for one or two dimentions
            n_features = input_data.shape[-1]

            if input_data.ndim == 1:
                input_data = input_data.reshape((1, n_features))

            data = zeros(input_data.shape)
            for _ in range(self.n_nodes):
                data[:, random.randint(0, n_features - 1)] += 1

            input_data = multiply(input_data, data)

        predicted = signum(input_data.dot(self.weight),
                           upper_value=1, lower_value=0)
        return predicted.astype(int)



from neuralpy.core.properties import NonNegativeIntProperty
from .base import SingleStep


__all__ = ('SimpleStepMinimization',)


class SimpleStepMinimization(SingleStep):
    """ Algorithm Monotonicly minimize learning step on each iteration.
    Probably this is most simple step minimization idea.

    Parameters
    ----------
    epochs_step_minimizator : int
        The parameter controls the frequency reduction step with respect
        to epochs. Defaults to ``100`` epochs. Can't be less than ``1``.
        Less value mean that step decrease faster.

    Attributes
    ----------
    {first_step}

    Warns
    -----
    {bp_depending}

    Examples
    --------
    >>> from neuralpy import algorithms
    >>>
    >>> bpnet = algorithms.Backpropagation(
    ...     (2, 4, 1),
    ...     step=0.1,
    ...     verbose=False,
    ...     optimizations=[algorithms.SimpleStepMinimization]
    ... )
    >>>

    See Also
    --------
    :network:`SearchThenConverge`
    """
    epochs_step_minimizator = NonNegativeIntProperty(min_size=1, default=100)

    def after_weight_update(self, input_train, target_train):
        super(SimpleStepMinimization, self).after_weight_update(
            input_train, target_train
        )
        self.step = self.first_step / (
            1 + self.epoch / self.epochs_step_minimizator
        )



from numpy import multiply, concatenate

from neuralpy import algorithms
from neuralpy.layers import SoftmaxLayer, OutputLayer
from neuralpy.functions import errors
from .base import BaseEnsemble


__all__ = ('MixtureOfExperts',)


class MixtureOfExperts(BaseEnsemble):
    """ Mixture of Experts ensemble algorithm for Backpropagation
    based Neural Networks.

    Parameters
    ----------
    networks : list
        List of networks based on :network:`Backpropagation` algorithm.
    gating_network : object
        2 Layer Neural Network based on :network:`Backpropagation` which
        has :layer:`SoftmaxLayer` as first one and the last one must be
        the :layer:`OutputLayer`. Output layer size must be equal to number
        of networks in model. Also important to say that in every network
        input size must be equal.

    Methods
    -------
    {supervised_train_epochs}
    {full_methods}

    Examples
    --------
    >>> import numpy as np
    >>> from sklearn import datasets, preprocessing
    >>> from sklearn.cross_validation import train_test_split
    >>> from neuralpy import ensemble, algorithms, layers
    >>> from neuralpy.functions import rmsle
    >>>
    >>> np.random.seed(100)
    >>>
    >>> dataset = datasets.load_diabetes()
    >>> data, target = dataset.data, dataset.target
    >>> input_scaler = preprocessing.MinMaxScaler((-1 ,1))
    >>> output_scaler = preprocessing.MinMaxScaler()
    >>>
    >>> x_train, x_test, y_train, y_test = train_test_split(
    ...     input_scaler.fit_transform(data),
    ...     output_scaler.fit_transform(target),
    ...     train_size=0.8
    ... )
    ...
    >>>
    >>> insize, outsize = (10, 1)
    >>> networks = [
    ...     algorithms.Backpropagation((insize, 20, outsize), step=0.1,
    ...                                 verbose=False),
    ...     algorithms.Backpropagation((insize, 20, outsize), step=0.1,
    ...                                 verbose=False),
    ... ]
    >>> n_networks = len(networks)
    >>>
    >>> moe = ensemble.MixtureOfExperts(
    ...     networks=networks,
    ...     gating_network=algorithms.Backpropagation(
    ...         layers.SoftmaxLayer(insize) > layers.OutputLayer(n_networks),
    ...         step=0.1,
    ...         verbose=False,
    ...     )
    ... )
    ...
    >>> moe.train(x_train, y_train, epochs=300)
    >>> y_predicted = moe.predict(x_test)
    >>>
    >>> rmsle(output_scaler.inverse_transform(y_test),
    ...       output_scaler.inverse_transform(y_predicted).round())
    0.44680253132714459
    """
    def __init__(self, networks, gating_network=None):
        super(MixtureOfExperts, self).__init__(networks)

        if not isinstance(gating_network, algorithms.Backpropagation):
            raise ValueError("Gating network must use Backpropagation "
                             "learning algorihtm")

        for network in self.networks:
            if not isinstance(network, algorithms.Backpropagation):
                raise ValueError(
                    "Network must use Backpropagation learning algorithm, "
                    "got {0}".format(network.__class__.__name__)
                )

            if network.output_layer.input_size != 1:
                raise ValueError("Network must contains one output unit, got "
                                 "{0}".format(network.output_layer.input_size))

            if network.error != errors.mse:
                raise ValueError(
                    "Use only Mean Square Error (MSE) function in network, "
                    "got {0}".format(network.error.__name__)
                )

            network.verbose = False

        if gating_network.input_layer.__class__ != SoftmaxLayer:
            raise ValueError(
                "Input layer must be `SoftmaxLayer`, got `{0}`".format(
                    gating_network.input_layer.__class__.__name__
                )
            )

        if gating_network.output_layer.__class__ != OutputLayer:
            raise ValueError(
                "First layer must be `OutputLayer`, got `{0}`".format(
                    gating_network.output_layer.__class__.__name__
                )
            )

        gating_network.verbose = False
        gating_network_output_size = gating_network.output_layer.input_size
        n_networks = len(self.networks)

        if gating_network_output_size != n_networks:
            raise ValueError("Invalid Gating network output size. Expected "
                             "{0}, got {1}".format(n_networks,
                                                   gating_network_output_size))

        if gating_network.error != errors.mse:
            raise ValueError(
                "Use only Mean Square Error (MSE) function in network, "
                "got {0}".format(gating_network.error.__name__)
            )

        self.gating_network = gating_network

    def train(self, input_data, target_data, epochs=100):
        if target_data.ndim == 1:
            target_data = target_data.reshape((target_data.size, 1))

        output_size = target_data.shape[1]
        if output_size != 1:
            raise ValueError("Target data must contains only 1 column, got "
                             "{0}".format(output_size))

        input_size = input_data.shape[1]

        gating_network = self.gating_network
        gating_network_input_size = gating_network.input_layer.input_size

        if gating_network_input_size != input_size:
            raise ValueError(
                "Gating Network expected get {0} input features, got "
                "{1}".format(gating_network_input_size, input_size)
            )

        for epoch in range(epochs):
            probs = self.gating_network.predict(input_data)
            total_output = 0
            outputs = []

            for i, network in enumerate(self.networks):
                output = network.predict(input_data)
                outputs.append(output)
                total_output += multiply(output, probs[:, i:i + 1])

            outputs = concatenate(outputs, axis=1)

            for i, network in enumerate(self.networks):
                # This is simple solution for error derivative update
                # g * (expected - actual) = g * expected - g * actual
                # It's simple hack and probably could broke other
                # algorithms, but now it works fine, but I need
                # change it later
                weight_delta = network.get_weight_delta(
                    total_output * probs[:, i:i + 1],
                    target_data * probs[:, i:i + 1]
                )
                network.update_weights(weight_delta)
                network.after_weight_update(input_data, target_data)

            # The same as at comment above
            weight_delta = gating_network.get_weight_delta(
                total_output * probs,
                target_data * probs
            )
            gating_network.update_weights(weight_delta)
            gating_network.after_weight_update(input_data, target_data)

    def predict(self, input_data):
        probs = self.gating_network.predict(input_data)
        total_output = 0

        for i, network in enumerate(self.networks):
            output = network.predict(input_data)
            total_output += multiply(output, probs[:, i:i + 1])

        return total_output

    def __repr__(self):
        indent = ' ' * 4
        return (
            "{classname}(networks=[\n"
            "{double_indent}{networks}\n"
            "{indent}],\n"
            "{indent}gating_network={gating_network}\n"
            ")"
        ).format(
            classname=self.__class__.__name__,
            networks=',\n        '.join(map(repr, self.networks)),
            gating_network=repr(self.gating_network),
            indent=indent,
            double_indent=(2 * indent)
        )



from abc import abstractmethod

from neuralpy.core.base import BaseSkeleton
from neuralpy.core.properties import BoolProperty
from neuralpy.core.config import ConfigurableWithABC


__all__ = ('BaseEnsemble',)


class BaseEnsemble(BaseSkeleton, ConfigurableWithABC):
    """ Base class for ensemlbe algorithms.
    """
    shuffle_data = BoolProperty(default=False)

    def __init__(self, networks):
        self.networks = networks

        if len(self.networks) < 2:
            raise ValueError("Ensemble must contains at least 2 networks")

    @abstractmethod
    def train(self, input_data, target_data, *args, **kwargs):
        pass

    @abstractmethod
    def predict(self, input_data):
        pass

    def __repr__(self):
        return "{classname}(networks=[\n    {networks}\n])".format(
            classname=self.__class__.__name__,
            networks=',\n    '.join(map(repr, self.networks))
        )



from neuralpy.core.properties import (NonNegativeNumberProperty,
                                      BetweenZeroAndOneProperty)
from .base import SingleStep


__all__ = ('ErrorDifferenceStepUpdate',)


class ErrorDifferenceStepUpdate(SingleStep):
    """ This algorithm make step update base on error difference between
    epochs.

    Parameters
    ----------
    update_for_smaller_error : float
        Multiplies this option to ``step`` in if the error was less than in
        previous epochs. Defaults to ``1.05``. Value can't be less
        than ``1``.
    update_for_bigger_error : float
        Multiplies this option to ``step`` in if the error was more than in
        previous epochs. Defaults to ``0.7``.
    error_difference : float
        The value indicates how many had to increase the error from the
        previous epochs that would produce a reduction step. Defaults
        to ``1.04``. Value can't be less than ``1``.

    Attributes
    ----------
    {first_step}

    Warns
    -----
    {bp_depending}

    Examples
    --------
    >>> from neuralpy import algorithms
    >>>
    >>> bpnet = algorithms.Backpropagation(
    ...     (2, 4, 1),
    ...     step=0.1,
    ...     verbose=False,
    ...     optimizations=[algorithms.ErrorDifferenceStepUpdate]
    ... )
    >>>
    """
    update_for_smaller_error = NonNegativeNumberProperty(default=1.05,
                                                         min_size=1)
    update_for_bigger_error = BetweenZeroAndOneProperty(default=0.7)
    error_difference = NonNegativeNumberProperty(default=1.04, min_size=1)

    def new_step(self):
        current_step = self.step

        if not self.errors_in:
            return current_step

        last_error = self.last_error_in()
        previous_error = self.previous_error()

        if previous_error is None:
            return current_step

        elif last_error < previous_error:
            return self.update_for_smaller_error * current_step

        elif last_error >= self.error_difference * previous_error:
            return self.update_for_bigger_error * current_step

        return current_step

    def after_weight_update(self, input_train, target_train):
        super(ErrorDifferenceStepUpdate, self).after_weight_update(
            input_train, target_train
        )
        self.step = self.new_step()



from numpy import zeros

from neuralpy.functions import signum
from .utils import sign2bin, bin2sign
from .base import DiscreteMemory


__all__ = ('DiscreteBAM',)


class DiscreteBAM(DiscreteMemory):
    """ Discrete BAM Network with associations.
    Network associate every input with some target value.

    Notes
    -----
    *{discrete_data_note}

    Methods
    -------
    train(input_data, output_data)
        Train network and update network weights.
    predict_output(input_data)
        Based on input date network predict it output.
    predict(input_data)
        The same as ``predict_output``
    predict_input(output_data)
        Based on output date network predict it input.

    Examples
    --------
    >>> import numpy as np
    >>> from neuralpy import algorithms
    >>>
    >>> zero = np.matrix([
    ...     0, 1, 1, 1, 0,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     0, 1, 1, 1, 0
    ... ])
    >>>
    >>> one = np.matrix([
    ...     0, 1, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0
    ... ])
    >>> zero.reshape((5, 6))
    >>>
    >>> half_zero = np.matrix([
    ...     0, 1, 1, 1, 0,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     0, 0, 0, 0, 0,
    ...     0, 0, 0, 0, 0,
    ...     0, 0, 0, 0, 0,
    ... ])
    >>>
    >>> zero_hint = np.matrix([[0, 1, 0, 0]])
    >>> one_hint = np.matrix([[1, 0, 0, 0]])
    >>>
    >>> data = np.concatenate([zero, one], axis=0)
    >>> hints = np.concatenate([zero_hint, one_hint], axis=0)
    >>>
    >>> bamnet = algorithms.DiscreteBAM()
    >>> bamnet.train(data.copy(), hints.copy())
    >>>
    >>> bamnet.predict(half_zero)
    matrix([[0, 1, 0, 0]])
    >>> zero_hint
    matrix([[0, 1, 0, 0]])
    """
    def __init__(self, **options):
        super(DiscreteBAM, self).__init__(**options)
        self.weight = None

    def format_predict(self, predicted_result):
        if self.weight is None:
            raise AttributeError("Train network before predict values")

        lower_value, upper_value = (-1, 1)
        predicted = signum(predicted_result, upper_value, lower_value)

        return sign2bin(predicted.astype(int))

    def predict_input(self, output_data):
        output_data = bin2sign(output_data)
        return self.format_predict(output_data.dot(self.weight.T))

    def predict_output(self, input_data):
        input_data = bin2sign(input_data)
        return self.format_predict(input_data.dot(self.weight))

    def train(self, input_data, output_data):
        self.discrete_validation(input_data)
        self.discrete_validation(output_data)

        output_data = bin2sign(output_data)
        input_data = bin2sign(input_data)

        _, wight_nrows = input_data.shape
        _, wight_ncols = output_data.shape
        weight_shape = (wight_nrows, wight_ncols)

        if self.weight is None:
            self.weight = zeros(weight_shape)

        if self.weight.shape != weight_shape:
            raise ValueError("Invalid input shapes. Number of input "
                             "features must be equal to {} and output "
                             "features - {}".format(wight_nrows, wight_ncols))

        self.weight += input_data.T.dot(output_data)

    predict = predict_output



from neuralpy.algorithms.feedforward import FeedForwardNetwork
from neuralpy.network.learning import SupervisedLearning
from neuralpy.functions import linear_error
from neuralpy.layers import StepLayer, OutputLayer


__all__ = ('SimpleTwoLayerNetwork',)


class SimpleTwoLayerNetwork(SupervisedLearning, FeedForwardNetwork):
    """ Base class for feedforward neural network without hidden layers.
    Input layer is always :layer:`StepLayer`.
    """
    def __init__(self, layer_sizes, **options):
        if len(layer_sizes) != 2:
            raise ValueError("This network must contains two layers.")

        input_layer_size, output_layer_size = layer_sizes

        super(SimpleTwoLayerNetwork, self).__init__(
            StepLayer(input_layer_size) > OutputLayer(output_layer_size),
            **options
        )

        self.input_layer = self.layers[0]
        self.error = linear_error

    def setup_defaults(self):
        del self.use_raw_predict_at_error

    def raw_predict(self, input_data):
        input_layer = self.input_layer

        input_data = input_layer.preformat_input(input_data)

        self.input_data = input_data
        self.summated = input_layer.summator(input_data)

        return input_layer.activation_function(self.summated)

    def update_weights(self, weight_deltas):
        self.input_layer.weight += self.step * weight_deltas



import numpy as np

from neuralpy.core.properties import NumberProperty
from .backpropagation import Backpropagation


__all__ = ('MinibatchGradientDescent',)


class MinibatchGradientDescent(Backpropagation):
    """ Mini-batch Gradient Descent algorithm.

    Parameters
    ----------
    batch_size : int
        Setup batch size for learning process, defaults to ``10``.
    {optimizations}
    {raw_predict_param}
    {full_params}

    Methods
    -------
    {supervised_train}
    {full_methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neuralpy import algorithms
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> mgdnet = algorithms.MinibatchGradientDescent(
    ...     (2, 3, 1),
    ...     verbose=False,
    ...     batch_size=1
    ... )
    >>> mgdnet.train(x_train, y_train)

    See Also
    --------
    :network:`Backpropagation` : Backpropagation algorithm.
    """
    batch_size = NumberProperty(default=10)

    def iter_batches(self, input_train, target_train):
        count_of_data = input_train.shape[0]

        for i in range(0, count_of_data, self.batch_size):
            batch = slice(i, i + self.batch_size)
            yield input_train[batch], target_train[batch]

    def train_batch(self, input_data, target_data):
        deltas = [np.zeros(l.weight.shape) for l in self.train_layers]

        predict_for_error = self.predict_for_error
        get_gradient = self.get_gradient

        for input_row, target_row in zip(input_data, target_data):
            output_train = predict_for_error(input_data)

            self.output_train = output_train

            weight_delta = get_gradient(output_train, target_data)
            deltas = map(sum, zip(deltas, weight_delta))

        self.weight_delta = [-delta / len(input_data) for delta in deltas]
        return self.weight_delta

    def train_epoch(self, input_train, target_train):
        batches = self.iter_batches(input_train, target_train)

        train_batch = self.train_batch
        update_weights = self.update_weights
        after_weight_update = self.after_weight_update

        for input_data, target_data in batches:
            weight_delta = train_batch(input_data, target_data)
            update_weights(weight_delta)
            after_weight_update(input_train, target_train)

        return self.error(self.predict_for_error(input_train), target_train)



from neuralpy.core.properties import NonNegativeNumberProperty
from .base import BaseStepAssociative


__all__ = ('HebbRule',)


class HebbRule(BaseStepAssociative):
    """ Hebbian Learning Unsupervised Neural Network.
    Network can learn associations from data and emulate similar behaviour
    as dog in Pavlov experiment.

    Notes
    -----
    * First layer must always be :layer:`StepLayer`.
    * Network must contains exacly 2 layers.
    * Network always generate weights which contains ``0`` weight for \
    conditioned stimulus and ``1`` otherwise. This setup helps you controll \
    your default state for learning features. Other type of weight you can \
    setup as optional parameter ``weight`` in input layer.
    * No bias.

    Parameters
    ----------
    decay_rate : float
        Decay rate is control your network weights. It helps network
        'forgote' information and control weight sizes. Without this
        parameter network weight will grow. Defaults to ``0.2``.
    n_unconditioned : int
        This value control number of features which are unconditioned
        stimulus for network. Defaults to ``1``. Can be any integer value
        bigger than ``1``, but less than feature space.
    {step}
    {show_epoch}
    {shuffle_data}
    {error}
    {verbose}
    {full_signals}

    Methods
    -------
    {unsupervised_train_epochs}
    {full_methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neuralpy import algorithms, layers
    >>>
    ... pavlov_dog_data = np.array([
    ...     [1, 0],  # food, no bell
    ...     [1, 1],  # food, bell
    ... ])
    >>> dog_test_cases = np.array([
    ...     [0, 0],  # no food, no bell
    ...     [0, 1],  # no food, bell
    ...     [1, 0],  # food, no bell
    ...     [1, 1],  # food, bell
    ... ])
    >>>
    >>> hebbnet = algorithms.HebbRule(
    ...     layers.StepLayer(2) > layers.OutputLayer(1),
    ...     n_unconditioned=1,
    ...     step=0.1,
    ...     decay_rate=0.8,
    ...     verbose=False
    ... )
    >>> hebbnet.train(pavlov_dog_data, epochs=2)
    >>> hebbnet.predict(dog_test_cases)
    array([[-1],
           [ 1],
           [ 1],
           [ 1]])
    """
    decay_rate = NonNegativeNumberProperty(default=0.2)

    def init_layers(self):
        layer_default_weight = self.input_layer.weight
        super(HebbRule, self).init_layers()

        if layer_default_weight is None:
            input_layer = self.input_layer
            unconditioned = self.n_unconditioned

            input_layer.weight[unconditioned:, :] = 0
            input_layer.weight[:unconditioned, :] = 1

    def weight_delta(self, input_row, layer_output):
        unconditioned = self.n_unconditioned
        update_from_column = unconditioned - self.use_bias
        weight = self.input_layer.weight[unconditioned:, :]
        delta = input_row[:, update_from_column:].T.dot(layer_output)
        return -self.decay_rate * weight + self.step * delta



from .backprop.backpropagation import *
from .backprop.levenberg_marquardt import *
from .backprop.gradient_descent import *
from .backprop.quasi_newton import *
from .backprop.conjugate_gradient import *
from .backprop.hessian_diagonal import *
from .backprop.rprop import *
from .backprop.quickprop import *
from .backprop.momentum import *

from .backprop.weights.weight_decay import *
from .backprop.weights.weight_elimination import *

from .backprop.steps.simple_step_minimization import *
from .backprop.steps.search_then_converge import *
from .backprop.steps.error_difference_update import *
from .backprop.steps.leak_step import *
from .backprop.steps.linear_search import *
from .backprop.steps.wolfe_search import *

from .memory.discrete_hopfield_network import *
from .memory.bam import *
from .memory.cmac import *

from .associative.oja import *
from .associative.hebb import *
from .associative.instar import *
from .associative.kohonen import *

from .competitive.sofm import *
from .competitive.art import *

from .rbfn.pnn import *
from .rbfn.rbf_kmeans import *
from .rbfn.grnn import *

from .basics.lms import *
from .basics.modify_relaxation import *
from .basics.perceptron import *



import numpy as np
import matplotlib.pyplot as plt

from neuralpy import algorithms, layers


np.random.seed(0)


input_data = np.array([
    [0.1961, 0.9806],
    [-0.1961, 0.9806],
    [0.9806, 0.1961],
    [0.9806, -0.1961],
    [-0.5812, -0.8137],
    [-0.8137, -0.5812],
])


input_layer = layers.LinearLayer(2)
output_layer = layers.CompetitiveOutputLayer(3)

sofmnet = algorithms.SOFM(
    # Connection
    input_layer > output_layer,
    # SOFM settingsp
    learning_radius=0,
    features_grid=(3, 1),
    # Network settings
    use_bias=False,
    step=0.1,
    show_epoch=100,
    shuffle_data=True,
)

plt.plot(input_data.T[0:1, :], input_data.T[1:2, :], 'ko')
sofmnet.train(input_data, epochs=100)

print("> Draw plot")
plt.xlim(-1, 1.2)
plt.ylim(-1, 1.2)

plt.plot(
    sofmnet.input_layer.weight[0:1, :],
    sofmnet.input_layer.weight[1:2, :],
    'bx'
)
plt.show()

for data in input_data:
    print(sofmnet.predict(np.reshape(data, (2, 1)).T))



from abc import abstractmethod

from neuralpy.core.config import ConfigurableWithABC


__all__ = ('Regression', 'Classification', 'Clustering')


class Regression(object):
    """ Mixin for regression.
    """


class Classification(ConfigurableWithABC):
    """ Mixin for classification.
    """
    @abstractmethod
    def predict_prob(self):
        """ Predict probabilities.
        """


class Clustering(object):
    """ Mixin for clustering.
    """



# -*- coding: utf-8 -*-
import math
from random import choice, randint

import numpy as np

from neuralpy.network.utils import shuffle


__all__ = ('make_reber', 'is_valid_by_reber', 'make_reber_classification')


avaliable_letters = 'TVPXS'
reber_rules = {
    0: [('T', 1), ('V', 2)],
    1: [('P', 1), ('T', 3)],
    2: [('X', 2), ('V', 4)],
    3: [('X', 2), ('S', None)],
    4: [('P', 3), ('S', None)],
}


def is_valid_by_reber(word):
    """ Сhecks whether a word belongs to grammar Reber.

    Parameters
    ----------
    word : str or list of letters
        The word that you want to test.

    Returns
    -------
    bool
        `True` if word valid by Reber grammar, `False` otherwise.

    Examples
    --------
    >>> from neuralpy.datasets import is_valid_by_reber
    >>>
    >>> is_valid_by_reber('TTS')
    True
    >>> is_valid_by_reber('STS')
    False
    >>>
    >>> is_valid_by_reber(['T', 'T', 'S'])
    True
    >>> is_valid_by_reber(['S', 'T', 'S'])
    False
    """
    if not word.endswith("S"):
        return False

    position = 0
    for letter in word:
        possible_letters = reber_rules[position]
        letters = [step[0] for step in possible_letters]
        if letter not in letters:
            return False
        _, position = possible_letters[letters.index(letter)]
    return True


def make_reber(n_words=100):
    """ Generate list of Reber grammar words.

    Parameters
    ----------
    n_words : int
        Number of reber words, defaults to `100`.

    Returns
    -------
    list
        List of Reber words.

    Examples
    --------
    >>> from neuralpy.datasets import make_reber
    >>>
    >>> make_reber(4)
    ['TPTXVS', 'VXXVS', 'TPPTS', 'TTXVPXXVS']
    """
    if n_words < 1:
        raise ValueError("Must be at least one word")

    words = []
    for i in range(n_words):
        position = 0
        word = []

        while position is not None:
            possible_letters = reber_rules[position]
            letter, position = choice(possible_letters)
            word.append(letter)

        words.append(''.join(word))
    return words


def make_reber_classification(n_samples, invalid_size=0.5):
    """ Generate random dataset for Reber grammar classification.
    Invalid words contains the same letters as at Reber grammar, but
    they are build whithout grammar rules.

    Parameters
    ----------
    n_samples : int
        Number of samples in dataset.
    invalid_size : float
        Proportion of invalid words in dataset, defaults to `0.5`. Value
        must be between 0 and 1, but not equal to them.

    Returns
    -------
    tuple
        Return two lists. First contains words and second - labels for them.

    Examples
    --------
    >>> from neuralpy.datasets import make_reber_classification
    >>>
    >>> data, labels = make_reber_classification(10, invalid_size=0.5)
    >>> data
    array(['SXSXVSXXVX', 'VVPS', 'VVPSXTTS', 'VVS', 'VXVS', 'VVS',
           'PPTTTXPSPTV', 'VTTSXVPTXVXT', 'VSSXSTX', 'TTXVS'],
          dtype='<U12')
    >>> labels
    array([0, 1, 0, 1, 1, 1, 0, 0, 0, 1])
    """
    if n_samples < 2:
        raise ValueError("There are must be at least 2 samples.")

    if invalid_size <= 0 or invalid_size >= 1:
        raise ValueError("`invalid_size` property must be between zero and"
                         "one, but not equal.")

    n_valid_words = int(math.ceil(n_samples * invalid_size))
    n_invalid_words = n_samples - n_valid_words

    valid_words = make_reber(n_valid_words)
    valid_labels = [1] * n_valid_words

    invalid_words = []
    invalid_labels = [0] * n_valid_words

    for i in range(n_invalid_words):
        word_length = randint(3, 14)
        word = [choice(avaliable_letters) for _ in range(word_length)]
        invalid_words.append(''.join(word))

    return shuffle(
        np.array(valid_words + invalid_words),
        np.array(valid_labels + invalid_labels)
    )



import numpy as np
from sklearn import datasets, preprocessing
from sklearn.cross_validation import train_test_split
from neuralpy import algorithms, layers
from neuralpy.functions import rmsle


np.random.seed(0)

dataset = datasets.load_boston()
data, target = dataset.data, dataset.target

data_scaler = preprocessing.MinMaxScaler()
target_scaler = preprocessing.MinMaxScaler()

data = data_scaler.fit_transform(data)
target = target_scaler.fit_transform(target)

x_train, x_test, y_train, y_test = train_test_split(
    data, target, train_size=0.85
)

cgnet = algorithms.ConjugateGradient(
    connection=[
        layers.SigmoidLayer(13),
        layers.SigmoidLayer(50),
        layers.OutputLayer(1),
    ],
    search_method='golden',
    show_epoch=25,
    optimizations=[algorithms.LinearSearch],
)

cgnet.train(x_train, y_train, x_test, y_test, epochs=100)
cgnet.plot_errors()

y_predict = cgnet.predict(x_test).round(1)
error = rmsle(target_scaler.inverse_transform(y_test),
              target_scaler.inverse_transform(y_predict))
print("RMSLE = {}".format(error))



from numpy import ones

from neuralpy.core.config import Configurable
from neuralpy.algorithms.backprop import LEARING_RATE_UPDATE


__all__ = ('SingleStep', 'MultiSteps')


class LearningRateConfigurable(Configurable):
    optimization_type = LEARING_RATE_UPDATE


class SingleStep(LearningRateConfigurable):
    """ Base class for backpropagation step algorithms which allow use single
    step for all layers.

    Attributes
    ----------
    {first_step}
    """
    def setup_defaults(self):
        self.first_step = self.step


class MultiSteps(LearningRateConfigurable):
    """ Base class for step algorithms which allow use unique step for
    every layer.

    Attributes
    ----------
    {steps}
    """
    def init_layers(self):
        super(MultiSteps, self).init_layers()
        steps = self.steps = []

        for layer in self.train_layers:
            steps.append(ones(layer.size) * self.step)

    def layer_step(self, layer_number):
        return self.steps[layer_number]



from __future__ import division
from numpy import log, sqrt, abs as np_abs, sum as np_sum

from neuralpy.functions import with_derivative


__all__ = ('mse', 'linear_error', 'cross_entropy_error', 'mae',
           'kullback_leibler', 'rmsle')


def _preformat_inputs(actual, predicted):
    if actual.ndim == 1:
        actual = actual.reshape((-1, 1))

    if predicted.ndim == 1:
        predicted = predicted.reshape((-1, 1))

    if actual.shape != predicted.shape:
        raise ValueError("Actual and predicted values have different shapes.")

    return actual, predicted


def linear_error(actual, predicted):
    """ Linear error.
    """
    actual, predicted = _preformat_inputs(actual, predicted)
    return predicted - actual


def mae(actual, predicted):
    """ Mean absolute error.
    """
    actual, predicted = _preformat_inputs(actual, predicted)
    data_size = actual.shape[0]
    return np_abs(predicted - actual).sum() / data_size


def mse_deriv(actual, predicted):
    """ Mean square error derivative.
    """
    actual, predicted = _preformat_inputs(actual, predicted)
    data_size = actual.shape[0]
    return (2 / data_size) * (actual - predicted)


@with_derivative(mse_deriv)
def mse(actual, predicted):
    """ Mean square error.
    """
    actual, predicted = _preformat_inputs(actual, predicted)
    data_size = actual.shape[0]
    return (1 / data_size) * np_sum((predicted - actual) ** 2)


def cross_entropy_error_deriv(actual, predicted, espilon=1e-10):
    """ Cross entropy error derivative.
    """
    actual, predicted = _preformat_inputs(actual, predicted)
    count_of_inputs = actual.shape[0]
    return (
        actual - predicted
    ) / (
        count_of_inputs * actual * (1 - actual) + espilon
    )


@with_derivative(cross_entropy_error_deriv)
def cross_entropy_error(actual, predicted, espilon=1e-10):
    """ Cross entropy error.
    """
    actual, predicted = _preformat_inputs(actual, predicted)
    count_of_inputs = actual.shape[0]
    return -(1 / count_of_inputs) * np_sum(
        (
            predicted * log(actual + espilon) +
            (1 - predicted) * log(1 - actual + espilon)
        )
    )


def kullback_leibler_deriv(actual, predicted):
    """ Kullback-Leibler error derivative.
    """
    actual, predicted = _preformat_inputs(actual, predicted)
    count_of_inputs = actual.shape[0]
    return (actual - predicted) / (
        count_of_inputs * actual * (1 - actual)
    )


@with_derivative(kullback_leibler_deriv)
def kullback_leibler(actual, predicted):
    """ Kullback-Leibler error.
    """
    actual, predicted = _preformat_inputs(actual, predicted)
    count_of_inputs = actual.shape[0]
    return (1. / count_of_inputs) * np_sum(
        predicted * log(predicted / actual) +
        (1 - predicted) * log((1 - predicted) / (1 - actual))
    )


def rmsle_deriv(actual, predicted):
    """ Root mean squared logarithmic error derivative.
    """
    actual, predicted = _preformat_inputs(actual, predicted)
    count_of = predicted.shape[0]

    logarithm_difference_deriv = log(
        (actual + 1) / (predicted + 1)
    )
    return logarithm_difference_deriv / (
        count_of * (actual + 1) * (rmsle(
            actual, predicted
        ))
    )


@with_derivative(rmsle_deriv)
def rmsle(actual, predicted):
    """ Root mean squared logarithmic error.
    """
    actual, predicted = _preformat_inputs(actual, predicted)
    count_of = predicted.shape[0]
    square_logarithm_difference = log((actual + 1) / (predicted + 1)) ** 2
    return sqrt((1 / count_of) * np_sum(square_logarithm_difference))



import numpy as np
import matplotlib.pyplot as plt


# Got color from page:
# http://matplotlib.org/examples/pylab_examples/custom_cmap.html
blue_red_cmap = {
    'red':  ((0.00, 0.0, 0.0),
             (0.25, 0.0, 0.0),
             (0.50, 0.8, 1.0),
             (0.75, 1.0, 1.0),
             (1.00, 0.4, 1.0)),

    'green': ((0.00, 0.0, 0.0),
              (0.25, 0.0, 0.0),
              (0.50, 0.9, 0.9),
              (0.75, 0.0, 0.0),
              (1.00, 0.0, 0.0)),

    'blue':  ((0.00, 0.0, 0.4),
              (0.25, 1.0, 1.0),
              (0.50, 1.0, 0.8),
              (0.75, 0.0, 0.0),
              (1.00, 0.0, 0.0))
}


def draw_countour(xgrid, ygrid, target_function):
    output = np.zeros((xgrid.shape[0], ygrid.shape[0]))

    for i, x in enumerate(xgrid):
        for j, y in enumerate(ygrid):
            output[j, i] = target_function(x, y)

    X, Y = np.meshgrid(xgrid, ygrid)

    plt.register_cmap(name='BlueRed', data=blue_red_cmap)
    plt.contourf(X, Y, output, 50, alpha=.75, cmap='BlueRed')
    plt.colorbar()


def weight_quiver(weights, color='c'):
    plt.quiver(weights[0, :-1],
               weights[1, :-1],
               weights[0, 1:] - weights[0, :-1],
               weights[1, 1:] - weights[1, :-1],
               scale_units='xy', angles='xy', scale=1,
               color=color)



import copy

from numpy import where, sign, ones, clip

from neuralpy.core.properties import (NonNegativeNumberProperty,
                                      BetweenZeroAndOneProperty)
from .backpropagation import Backpropagation


__all__ = ('RPROP', 'IRPROPPlus')


class RPROP(Backpropagation):
    """ RPROP :network:`Backpropagation` algorithm optimization.

    Parameters
    ----------
    {rprop_params}
    {optimizations}
    {raw_predict_param}
    {full_params}

    Methods
    -------
    {supervised_train}
    {full_methods}

    Examples
    --------
    Simple example

    >>> import numpy as np
    >>> from neuralpy import algorithms
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> rpropnet = algorithms.RPROP(
    ...     (2, 3, 1),
    ...     verbose=False
    ... )
    >>> rpropnet.train(x_train, y_train)

    See Also
    --------
    :network:`IRPROPPlus` : iRPROP+ algorithm.
    :network:`Backpropagation` : Backpropagation algorithm.
    """

    __rprop_params = """minimum_step : float
        Minimum possible value for step. Defaults to ``0.1``.
    maximum_step : float
        Maximum possible value for step. Defaults to ``50``.
    increase_factor : float
        Increase factor for step in case when gradient doesn't change
        sign compare to previous epoch.
    decrease_factor : float
        Decrease factor for step in case when gradient changes sign
        compare to previous epoch.
    """

    shared_docs = {"rprop_params": __rprop_params}

    # This properties correct upper and lower bounds for steps.
    minimum_step = NonNegativeNumberProperty(default=0.1)
    maximum_step = NonNegativeNumberProperty(default=50)

    # This properties increase/decrease step by deviding it to
    # some coeffitient.
    increase_factor = NonNegativeNumberProperty(min_size=1, default=1.2)
    decrease_factor = BetweenZeroAndOneProperty(default=0.5)

    def init_layers(self):
        super(RPROP, self).init_layers()
        steps = self.steps = []

        for layer in self.train_layers:
            steps.append(ones(layer.size) * self.step)

    def get_flip_sign_weight_delta(self, layer_number):
        return self.prev_weight_deltas[layer_number]

    def layer_weight_update(self, delta, layer_number):
        if not hasattr(self, 'prev_gradients'):
            prev_gradient = 0
            prev_weight_delta = 0
        else:
            prev_gradient = self.prev_gradients[layer_number]
            prev_weight_delta = self.get_flip_sign_weight_delta(layer_number)

        step = self.steps[layer_number]
        gradient = self.gradients[layer_number]

        grad_product = prev_gradient * gradient
        negative_gradients = grad_product < 0

        step = self.steps[layer_number] = clip(
            where(
                grad_product > 0,
                # Increase step for gradients which switch signs
                step * self.increase_factor,
                where(
                    negative_gradients,
                    # Decrease step for gradients whcih switch signs
                    step * self.decrease_factor,
                    # Setup the same step value
                    step
                )
            ),
            self.minimum_step,
            self.maximum_step,
        )

        output = where(
            negative_gradients,
            -prev_weight_delta,
            -sign(gradient) * step
        )
        gradient[negative_gradients] = 0

        self.weight_deltas.append(output.copy())

        return output

    def update_weights(self, weight_deltas):
        self.weight_deltas = []
        super(RPROP, self).update_weights(weight_deltas)

        self.prev_weight_deltas = copy.copy(self.weight_deltas)
        self.prev_gradients = copy.copy(self.gradients)
        self.prev_steps = copy.copy(self.steps)


class IRPROPPlus(RPROP):
    """ iRPROP+ :network:`Backpropagation` algorithm optimization.

    Parameters
    ----------
    {rprop_params}
    {optimizations}
    {full_params}

    Methods
    -------
    {supervised_train}
    {full_methods}

    Examples
    --------
    Simple example

    >>> import numpy as np
    >>> from neuralpy import algorithms
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> rpropnet = algorithms.IRPROPPlus(
    ...     (2, 3, 1),
    ...     verbose=False
    ... )
    >>> rpropnet.train(x_train, y_train)

    See Also
    --------
    :network:`RPROP` : RPROP algorithm.
    :network:`Backpropagation` : Backpropagation algorithm.
    """
    def get_flip_sign_weight_delta(self, layer_number):
        prev_error = self.previous_error()
        last_error = self.last_error()
        prev_weight_delta = 0

        if prev_error is None or last_error > prev_error:
            prev_weight_delta = self.prev_weight_deltas[layer_number]

        return prev_weight_delta



from numpy import eye, newaxis, sign, isinf, clip, inner, outer
from numpy.linalg import norm

from neuralpy.core.properties import (ChoiceProperty,
                                      NonNegativeNumberProperty,
                                      BetweenZeroAndOneProperty)
from neuralpy.algorithms.utils import (matrix_list_in_one_vector,
                                       vector_to_list_of_matrix)
from .steps.wolfe_search import WolfeSearch
from .backpropagation import Backpropagation


__all__ = ('QuasiNewton',)


def bfgs(inverse_hessian, weight_delta, gradient_delta, maxrho=1e4):
    ident_matrix = eye(inverse_hessian.shape[0], dtype=int)

    rho = (1. / gradient_delta.dot(weight_delta))

    if isinf(rho):
        rho = maxrho * sign(rho)

    param1 = ident_matrix - outer(weight_delta, gradient_delta) * rho
    param2 = ident_matrix - outer(gradient_delta, weight_delta) * rho
    param3 = rho * outer(weight_delta, weight_delta)

    return param1.dot(inverse_hessian).dot(param2) + param3


def dfp(inverse_hessian, weight_delta, gradient_delta, maxnum=1e5):
    gradient_delta_t = gradient_delta[newaxis, :]
    gradient_delta = gradient_delta[:, newaxis]
    weight_delta = weight_delta[:, newaxis]

    quasi_dot_gradient = inverse_hessian.dot(gradient_delta)

    param1 = (
        weight_delta.dot(weight_delta.T)
    ) / (
        gradient_delta_t.dot(weight_delta)
    )
    param2_numerator = clip(
        quasi_dot_gradient.dot(gradient_delta_t) * inverse_hessian,
        a_min=-maxnum,
        a_max=maxnum
    )
    param2_denominator = gradient_delta_t.dot(quasi_dot_gradient)
    param2 = param2_numerator / param2_denominator

    return inverse_hessian + param1 - param2


def psb(inverse_hessian, weight_delta, gradient_delta, **options):
    gradient_delta_t = gradient_delta.T
    param = weight_delta - inverse_hessian.dot(gradient_delta)

    devider = (1. / inner(gradient_delta, gradient_delta)).item(0)
    param1 = outer(param, gradient_delta) + outer(gradient_delta, param)
    param2 = (
        inner(gradient_delta, param)
    ).item(0) * outer(gradient_delta, gradient_delta_t)

    return inverse_hessian + param1 * devider - param2 * devider ** 2


def sr1(inverse_hessian, weight_delta, gradient_delta, epsilon=1e-8):
    param = weight_delta - inverse_hessian.dot(gradient_delta)
    denominator = inner(param, gradient_delta)

    if abs(denominator) < epsilon * norm(param) * norm(gradient_delta):
        return inverse_hessian

    return inverse_hessian + outer(param, param) / denominator


class QuasiNewton(Backpropagation):
    """ Quasi-Newton :network:`Backpropagation` algorithm optimization.

    Parameters
    ----------
    update_function : {{'bfgs', 'dfp', 'psb', 'sr1'}}
        Update function. Defaults to ``bfgs``.
    h0_scale : float
        Factor that scale indentity matrix H0 on the first
        iteration step. Defaults to ``1``.
    gradient_tol : float
        In the gradient less than this value algorithm will stop training
        procedure. Defaults to ``1e-5``.
    {optimizations}
    {raw_predict_param}
    {full_params}

    Methods
    -------
    {supervised_train}
    {full_methods}

    Examples
    --------
    Simple example

    >>> import numpy as np
    >>> from neuralpy import algorithms
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> qnnet = algorithms.QuasiNewton(
    ...     (2, 3, 1),
    ...     update_function='bfgs',
    ...     verbose=False
    ... )
    >>> qnnet.train(x_train, y_train)

    See Also
    --------
    :network:`Backpropagation` : Backpropagation algorithm.
    """
    update_function = ChoiceProperty(
        default='bfgs',
        choices={
            'bfgs': bfgs,
            'dfp': dfp,
            'psb': psb,
            'sr1': sr1,
        }
    )
    h0_scale = NonNegativeNumberProperty(default=1)
    gradient_tol = BetweenZeroAndOneProperty(default=1e-5)

    default_optimizations = [WolfeSearch]

    def get_weight_delta(self, output_train, target_train):
        gradients = self.get_gradient(output_train, target_train)
        gradient = matrix_list_in_one_vector(gradients)

        if norm(gradient) < self.gradient_tol:
            raise StopIteration("Gradient norm less than {}"
                                "".format(self.gradient_tol))

        train_layers = self.train_layers
        weight = matrix_list_in_one_vector(
            (layer.weight for layer in train_layers)
        )

        if hasattr(self, 'prev_gradient'):
            # In first epoch we didn't have previous weights and
            # gradients. For this reason we skip quasi coefitient
            # computation.
            inverse_hessian = self.update_function(
                self.prev_inverse_hessian,
                weight - self.prev_weight,
                gradient - self.prev_gradient
            )
        else:
            inverse_hessian = self.h0_scale * eye(weight.size, dtype=int)

        self.prev_weight = weight.copy()
        self.prev_gradient = gradient.copy()
        self.prev_inverse_hessian = inverse_hessian

        return vector_to_list_of_matrix(
            -inverse_hessian.dot(gradient),
            (layer.size for layer in train_layers)
        )



from numpy import any as np_any

from neuralpy.core.base import BaseSkeleton
from neuralpy.core.config import Configurable


__all__ = ('DiscreteMemory',)


class DiscreteMemory(BaseSkeleton, Configurable):
    """ Base class for discrete memory networks.

    Notes
    -----
    * {discrete_data_note}
    """
    __discrete_data_note = """ Input and output data must contains only \
    binary values which are equal to settings in `discrete_values` property.
    """
    shared_docs = {'discrete_data_note': __discrete_data_note}

    def __init__(self, *args, **kwargs):
        self.discrete_values = (0, 1)
        super(DiscreteMemory, self).__init__(*args, **kwargs)

    def discrete_validation(self, matrix):
        """ Validate discrete matrix.

        Parameters
        ----------
        matrix : array-like
            Matrix for validation.

        Returns
        -------
        bool
            Got ``True`` all ``matrix`` discrete values are in
            `discrete_values` list and `False` otherwise.
        """
        lower_value, upper_value = self.discrete_values

        if np_any((matrix != lower_value) & (matrix != upper_value)):
            raise ValueError("This network is descrete. This mean that you "
                             "can use data which contains {0} and {1} "
                             "values".format(lower_value, upper_value))



import numpy as np

from neuralpy.core.properties import NonNegativeNumberProperty
from neuralpy.algorithms.basics.base import SimpleTwoLayerNetwork


__all__ = ('ModifiedRelaxation',)


class ModifiedRelaxation(SimpleTwoLayerNetwork):
    """ Modified Relaxation Neural Network. Simple linear network. If the
    output value of the network received more than the set limit, the
    weight is updated in the same way as the :network:`LMS`, if less
    than the set value - the update will be in proportion to the
    expected result.

    Parameters
    ----------
    dead_zone_radius : float
        Indicates the line between stable outcome network output and
        weak, and depending on the result of doing different updates.
    {full_params}

    Methods
    -------
    {supervised_train}
    {full_methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neuralpy import algorithms
    >>>
    >>> input_data = np.array([[1, 0], [2, 2], [3, 3], [0, 0]])
    >>> target_data = np.array([[1], [-1], [-1], [1]])
    >>>
    >>> mrnet = algorithms.ModifiedRelaxation((2, 1), step=1, verbose=False)
    >>> mrnet.train(input_data, target_data, epochs=100)
    >>> mrnet.predict(np.array([[4, 4], [-1, -1]]))
    array([[-1],
           [ 1]])

    See Also
    --------
    :network:`LMS` : LMS Neural Network.
    """
    dead_zone_radius = NonNegativeNumberProperty(default=0.1)

    def get_weight_delta(self, output_train, target_train):
        input_data = self.input_data
        update = np.where(
            np.abs(self.summated) >= self.dead_zone_radius,
            self.error(output_train, target_train),
            target_train
        )
        minimized_input = input_data / np.linalg.norm(input_data) ** 2

        return np.dot(minimized_input.T, update)



from abc import ABCMeta, abstractmethod, abstractproperty

from six import with_metaclass


__all__ = ('LayerConnection', 'ReccurentConnection', 'Connection',
           'ChainConnection', 'FAKE_CONNECTION', 'NetworkConnectionError')


FAKE_CONNECTION = [None]


class NetworkConnectionError(Exception):
    pass


class ChainConnection(object):
    def __init__(self):
        self.connection = None

    def __gt__(self, other):
        return LayerConnection(self, other)

    def __and__(self, other):
        return ReccurentConnection(self, other)


class Connection(with_metaclass(ABCMeta, ChainConnection)):
    def __init__(self, left, right):
        super(Connection, self).__init__()

        self.left = left.connection or left
        self.right = right.connection or right

        self.left.connection = self
        self.right.connection = self

        self.connect()

    @abstractmethod
    def connect(self):
        pass

    @abstractproperty
    @property
    def left_layer(self):
        pass

    @abstractproperty
    @property
    def right_layer(self):
        pass

    def __len__(self):
        return len(list(self.__iter__()))

    def __iter__(self):
        if isinstance(self.left, Connection):
            for conn in self.left:
                yield conn
        else:
            yield self.left

        if isinstance(self.right, Connection):
            for conn in self.right:
                yield conn
        else:
            yield self.right

    def __repr__(self):
        layers_reprs = map(repr, self)
        return ' > '.join(layers_reprs)


class LayerConnection(Connection):
    def connect(self):
        self.left_layer.relate_to(self.right_layer)

    @property
    def left_layer(self):
        if isinstance(self.left, Connection):
            return self.left.right_layer
        return self.left

    @property
    def right_layer(self):
        if isinstance(self.right, Connection):
            return self.right.left_layer
        return self.right

    @property
    def input_layer(self):
        if isinstance(self.left, Connection):
            return self.left.input_layer
        return self.left

    @property
    def output_layer(self):
        if isinstance(self.right, Connection):
            return self.right.output_layer
        return self.right


class ReccurentConnection(Connection):
    pass



from numpy import where, zeros, reshape, round as np_round

from .base import BaseEnsemble


__all__ = ('DynamicallyAveragedNetwork',)


class DynamicallyAveragedNetwork(BaseEnsemble):
    """ Dynamically Averaged Network (DAN) weighted ensamble for binary
    classification problems.

    Notes
    -----
    * Every network must has 1 output and result must be between 0 and 1.

    Parameters
    ----------
    networks : list
        List of Neural Networks.

    Methods
    -------
    {supervised_train}
    {full_methods}

    Attributes
    ----------
    weights : ndarray, shape = [n_predictors, n_networks]
        After you get prediction you can also check weight which you
        will get to combine the result.

    Examples
    --------
    >>> import numpy as np
    >>> from sklearn import datasets, metrics
    >>> from sklearn.cross_validation import train_test_split
    >>> from neuralpy import ensemble, algorithms
    >>>
    >>> np.random.seed(50)
    >>>
    >>> data, target = datasets.make_classification(300, n_features=4,
    >>>                                             n_classes=2)
    >>> x_train, x_test, y_train, y_test = train_test_split(
    >>>     data, target, train_size=0.7
    >>> )
    >>>
    >>> dan = ensemble.DynamicallyAveragedNetwork([
    >>>     algorithms.RPROP((4, 10, 1), step=0.1, maximum_step=1,
    >>>                       verbose=False),
    >>>     algorithms.Backpropagation((4, 5, 1), step=0.1,
    >>>                                 verbose=False)
    >>> ])
    >>> dan.train(x_train, y_train, epochs=500)
    >>> y_predicted = dan.predict(x_test)
    >>>
    >>> metrics.accuracy_score(y_test, y_predicted)
    0.97777777777777775
    """
    def __init__(self, networks):
        super(DynamicallyAveragedNetwork, self).__init__(networks)
        self.weights = None

        for network in networks:
            output_layer_size = network.output_layer.input_size
            network.verbose = False

            if output_layer_size != 1:
                raise ValueError(
                    "Final layer at network `{}` must has 1 output, got "
                    "{}".format(self.__class__.__name__, output_layer_size)
                )

    def train(self, input_data, target_data, *args, **kwargs):
        for network in self.networks:
            network.train(input_data, target_data, *args, **kwargs)

    def raw_predict(self, input_data):
        number_of_inputs = input_data.shape[0]
        network_certainties = zeros((number_of_inputs, len(self.networks)))
        network_outputs = network_certainties.copy()

        for i, network in enumerate(self.networks):
            output = network.predict(input_data)
            minval, maxval = output.min(), output.max()

            if not (0 <= minval <= 1 and 0 <= maxval <= 1):
                raise ValueError(
                    "Netwrok output must be in range [0, 1]. Network output "
                    "was in range [{}, {}]".format(minval, maxval)
                )

            certainty = where(output > 0.5, output, 1 - output)

            network_certainties[:, i:i + 1] = certainty
            network_outputs[:, i:i + 1] = output

        total_output_sum = reshape(network_certainties.sum(axis=1),
                                   (number_of_inputs, 1))
        self.weights = network_certainties / total_output_sum
        return (self.weights * network_outputs).sum(axis=1)

    def predict(self, input_data):
        raw_output = self.raw_predict(input_data)
        return np_round(raw_output)



from neuralpy.core.properties import (BetweenZeroAndOneProperty,
                                      NonNegativeNumberProperty)
from .base import MultiSteps


__all__ = ('DeltaBarDelta',)


class DeltaBarDelta(MultiSteps):
    beta = BetweenZeroAndOneProperty(default=0.5)
    increase_factor = NonNegativeNumberProperty(default=0.1)
    decrease_factor = BetweenZeroAndOneProperty(default=0.9)



import os
import importlib

import baker


CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))

manager = baker.Baker()


for filename in os.listdir(CURRENT_DIR):
    if filename not in ('__init__.py', 'main.py') and filename.endswith('py'):
        module_name = filename.rsplit('.')[0]
        module = importlib.import_module(
            'neuralpy.commands.{}'.format(module_name)
        )
        manager.command(name=module_name)(module.run)


def main():
    manager.run()



from numpy import zeros
from numpy.linalg import norm

from neuralpy.core.properties import (BetweenZeroAndOneProperty,
                                      NonNegativeNumberProperty)
from .base import MultiSteps


__all__ = ('LeakStepAdaptation',)


class LeakStepAdaptation(MultiSteps):
    """ Leak Learning Rate Adaptation algorithm for step adaptation procedure
    in backpropagation algortihm. By default every layer has the same value
    as ``step`` parameter in network, but after first training epoch they
    must be different.

    Parameters
    ----------
    leak_size : float
        Leak size control ratio of update variable which combine weight
        deltas from previous epochs, defaults to ``0.5``.
    alpha : float
        The ``alpha`` is control total step update ratio (It's similar to
        step role in weight update procedure). Defaults to ``0.5``.
    beta : float
        This similar to ``alpha``, but it control ration only for update
        matrix norms. Defaults to ``0.5``.

    Attributes
    ----------
    {steps}

    Warns
    -----
    {bp_depending}

    Examples
    --------
    >>> from neuralpy import algorithms
    >>>
    >>> bpnet = algorithms.Backpropagation(
    ...     (2, 4, 1),
    ...     step=0.1,
    ...     verbose=False,
    ...     optimizations=[algorithms.LeakStepAdaptation]
    ... )
    >>>
    """
    leak_size = BetweenZeroAndOneProperty(default=0.5)
    alpha = NonNegativeNumberProperty(default=0.5)
    beta = NonNegativeNumberProperty(default=0.5)

    def init_layers(self):
        super(LeakStepAdaptation, self).init_layers()
        updates = self.updates = []

        for layer in self.train_layers:
            updates.append(zeros(layer.size))

    def after_weight_update(self, input_train, target_train):
        super(LeakStepAdaptation, self).after_weight_update(input_train,
                                                            target_train)
        alpha = self.alpha
        beta = self.beta
        leak_size = self.leak_size

        weight_delta = self.weight_delta
        steps = self.steps
        updates = self.updates

        for i, layer in enumerate(self.train_layers):
            step = steps[i]
            update = updates[i]

            updates[i] = (1 - leak_size) * update + (
                leak_size * weight_delta[i]
            )
            steps[i] += alpha * step * (beta * norm(updates[i]) - step)



from functools import partial

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches

from neuralpy import algorithms, layers
from plots import draw_countour, weight_quiver


# Settings
TOTAL_EPOCHS = 30

# Initialize test data
input_data = np.array([
    [0.9, 0.3],
    [0.5, 0.3],
    [0.2, 0.1],
    [0.7, 0.5],
    [0.1, 0.8],
    [0.1, 0.9],
])
target_data = np.array([
    [1],
    [1],
    [1],
    [0],
    [0],
    [0],
])
default_weight = np.array([[-4.], [-4.]])

weights = None


def save_weight_in_epoch(net):
    """ Signal processor which save weight update for every
    epoch.
    """
    global weights
    input_layer_weight = net.train_layers[0].weight.copy()
    weights[:, net.epoch + 1:net.epoch + 2] = input_layer_weight


def get_connection():
    """ Generate new connections every time when we call it """
    input_layer = layers.SigmoidLayer(2, weight=default_weight.copy())
    output_layer = layers.OutputLayer(1)
    return input_layer > output_layer


# Setup default networks settings which we will use in all algorithms
# which we will vizualize.
network_settings = dict(
    # Network
    step=0.3,
    use_bias=False,
    # Signals
    train_epoch_end_signal=save_weight_in_epoch,
    verbose=False,
)


def draw_quiver(network_class, name, color='r'):
    """ Train algorithm and draw quiver for every epoch update
    for this algorithm.
    """
    global weights

    bpn = network_class(get_connection(), **network_settings)
    # 1000 is an upper limit for all network epochs, later we
    # will fix it size
    weights = np.zeros((2, 1000))
    weights[:, 0:1] = default_weight.copy()
    bpn.train(input_data, target_data, epsilon=0.125)
    weights = weights[:, :bpn.epoch + 1]
    weight_quiver(weights, color=color)

    label = "{name} ({n} steps)".format(name=name, n=bpn.epoch)
    return mpatches.Patch(color=color, label=label)


def target_function(network, x, y):
    network.input_layer.weight = np.array([[x], [y]])
    predicted = network.predict(input_data)
    return network.error(predicted, target_data)


# Get data for countour plot
bp_network = algorithms.Backpropagation(get_connection(), **network_settings)
network_target_function = partial(target_function, bp_network)

plt.figure()
plt.title("Approximation function contour plot")
plt.xlabel("First weight")
plt.ylabel("Second weight")

draw_countour(
    np.linspace(-4.5, 4, 50),
    np.linspace(-4.5, 4, 50),
    network_target_function
)

cgnet_class = partial(algorithms.ConjugateGradient,
                      optimizations=[algorithms.LinearSearch])

algorithms = (
    (algorithms.Backpropagation, 'Gradient Descent', 'k'),
    (algorithms.Momentum, 'Momentum', 'm'),
    (algorithms.RPROP, 'RPROP', 'c'),
    (cgnet_class, 'Conjugate Gradient', 'y'),
)

patches = []
for network_params in algorithms:
    quiver_patch = draw_quiver(*network_params)
    patches.append(quiver_patch)

plt.legend(handles=patches)
plt.show()



import numpy as np

from neuralpy.algorithms.basics.base import SimpleTwoLayerNetwork


__all__ = ('Perceptron',)


class Perceptron(SimpleTwoLayerNetwork):
    """ Perceptron Neural Network. Simples linear model in Neural
    Networks.

    Parameters
    ----------
    {full_params}

    Methods
    -------
    {supervised_train}
    {full_methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neuralpy import algorithms
    >>>
    >>> input_data = np.array([[1, 0], [2, 2], [3, 3], [0, 0]])
    >>> target_data = np.array([[1], [-1], [-1], [1]])
    >>>
    >>> mrnet = algorithms.Perceptron((2, 1), step=0.4, verbose=False)
    >>> mrnet.train(input_data, target_data, epochs=30)
    >>> mrnet.predict(np.array([[4, 4], [-1, -1]]))
    array([[-1],
           [ 1]])
    """
    def get_weight_delta(self, output_train, target_train):
        error_result = self.error(output_train, target_train)
        return np.dot(self.input_data.T, error_result)



from numpy.random import randn, random
from numpy.linalg import svd

from neuralpy.network.connections import LayerConnection
from neuralpy.helpers import import_class


__all__ = ('generate_layers', 'random_orthogonal', 'random_bounded')


DEFAULT_LAYER_CLASS = "neuralpy.layers.SigmoidLayer"
DEFAULT_OUTPUT_LAYER_CLASS = "neuralpy.layers.OutputLayer"


def generate_layers(layers_sizes):
    """ Create from list of layer sizes basic linear network.

    Parameters
    ----------
    layers_sizes : list or tuple
        Ordered lsit of network connection structure.
    """

    if len(layers_sizes) < 2:
        raise ValueError("Network must contains at least 2 layers.")

    default_layer_class = import_class(DEFAULT_LAYER_CLASS)
    default_output_layer_class = import_class(DEFAULT_OUTPUT_LAYER_CLASS)

    output_layer_size = layers_sizes.pop()
    connection = default_output_layer_class(output_layer_size)

    for input_size in reversed(layers_sizes):
        left_layer = default_layer_class(input_size)
        connection = LayerConnection(left_layer, connection)

    return connection


def random_orthogonal(size):
    """ Build random orthogonal 2D matrix.

    Parameters
    ----------
    size : tuple
        Generated matrix shape.
    """
    if len(size) != 2:
        raise ValueError("Size must be a tuple which contains 2"
                         "integer values.")
    rand_matrix = randn(*size)
    u, _, v = svd(rand_matrix, full_matrices=False)
    ortho_base = u if size[0] > size[1] else v
    return ortho_base[:size[0], :size[1]]


def random_bounded(size, left_bound=0, right_bound=1):
    """ Generate uniform random matrix which values between bounds.
    """
    random_weight = random(size)
    return random_weight * (right_bound - left_bound) + left_bound



from functools import partial

from numpy import sum as np_sum, abs as np_abs


__all__ = ('with_derivative', 'deriv_partial', 'get_partial_for_func',
           'normilize_error_output', 'has_deriv')


def with_derivative(deriv_func):
    """ Decorator which add derivative property to function.

    Examples
    --------
    >>> from neuralpy.functions import with_derivative
    >>>
    >>> def polynomial_deriv(x):
    ...     return 2 * x + 1
    ...
    >>> @with_derivative(polynomial_deriv)
    ... def polynomial(x):
    ...     return x**2 + x + 1
    ...
    >>> polynomial(10)
    111
    >>> polynomial.deriv(10)
    21
    """
    def decorator(func):
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        wrapper.__name__ = func.__name__
        wrapper.deriv = deriv_func
        wrapper.func_desc = "'{}' with derivative function '{}'".format(
            func.__name__, deriv_func.__name__
        )

        return wrapper
    return decorator


def deriv_partial(func, *args, **kwargs):
    """ Override around `partial` function from `functools` library.
    This function will add the same partial arguments also for inner
    derivatives. Check all derivatives recursivly.

    Parameters
    ----------
    func : function
        Function which must override.
    *args
        list of arguments for function from parameter `func`.
    **kwargs
        dictionary of arguments for function from parameter `func`.

    Returns
    -------
    function
        The same function around `partial` function from default
        python library `functools`.
    """
    func_partial = get_partial_for_func(func.deriv)
    new_func = partial(func, *args, **kwargs)
    new_func.deriv = func_partial(func.deriv, *args, **kwargs)
    return new_func


def get_partial_for_func(function):
    """ Returns valid partial function.

    Parameters
    ----------
    function : function
        Function for which you need partial function.

    Returns
    -------
    function
        Return `partial` function from default python library `functools`
        if function hasn't derivative, and modified version otherwise.
    """
    return deriv_partial if has_deriv(function) else partial


def normilize_error_output(output):
    """ Normalize error output when result is non-scalar.

    Parameters
    ----------
    output : array-like
        Input can be any numpy array or matrix.

    Returns
    -------
    int, float
        Return sum of all absolute values.
    """
    return np_sum(np_abs(output))


def has_deriv(function):
    """ Check that function has derivative.

    Parameters
    ----------
    function : function
        Function which need to validate.

    Returns
    -------
    bool
        True if function has derivative, False otherwise.
    """
    return hasattr(function, 'deriv')



import os
import sys
import inspect
import importlib

from neuralpy.helpers.logs import green


sys.path.append(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), '..', '..')
)


FILES = (
    ('Algorithms', (
        ('algorithms',),
    )),
    ('Functions', (
        ('functions', 'errors'),
    )),
    ('Layers', (
        ('layers', 'layers'),
        ('layers', 'output'),
    ))
)


def print_with_indent(text, indent=0):
    print(' ' * indent, '-', text)


def show_module_classes(module):
    for object_name in dir(module):
        object_value = getattr(module, object_name)

        is_not_valid_name = object_name.startswith('__')

        if not (is_not_valid_name or inspect.ismodule(object_value)):
            yield object_name


def get_section(section_number):
    n_sections = len(FILES)

    if section_number > n_sections:
        raise ValueError(
            "There is no section number {}. You can check sections "
            "from 1 to {}".format(section_number, n_sections)
        )

    section_name, filepatterns = FILES[section_number - 1]
    print(green('{}. {}'.format(section_number, section_name)))

    for filepattern in filepatterns:
        module_name = '.'.join(['neuralpy'] + list(filepattern))

        print_with_indent(module_name, indent=2)
        module = importlib.import_module(module_name)

        if hasattr(module, '__all__'):
            special_objects = module.__all__
        else:
            special_objects = show_module_classes(module)

        for special_object in sorted(special_objects):
            print_with_indent(special_object, indent=4)

        print('')


def run(section=None):
    if section is not None:
        get_section(int(section))
    else:
        for i in range(1, len(FILES) + 1):
            get_section(i)



from numpy import unique, zeros, dot, sum as np_sum

from neuralpy import layers
from neuralpy.core.properties import NonNegativeNumberProperty
from neuralpy.network.base import BaseNetwork
from neuralpy.network.connections import FAKE_CONNECTION
from neuralpy.network.learning import LazyLearning
from neuralpy.network.types import Classification

from .utils import pdf_between_data


__all__ = ('PNN',)


class PNN(LazyLearning, Classification, BaseNetwork):
    """ Probabilistic Neural Network for classification.

    Parameters
    ----------
    standard_deviation : float
        standard deviation for PDF function, default to 0.1.
    {show_epoch}
    {shuffle_data}
    {full_signals}
    {verbose}

    Methods
    -------
    {supervised_train_lazy}
    {full_methods}

    Examples
    --------
    >>> import numpy as np
    >>>
    >>> from sklearn import datasets
    >>> from sklearn import metrics
    >>> from sklearn.cross_validation import train_test_split
    >>>
    >>> from neuralpy.algorithms import PNN
    >>>
    >>> np.random.seed(0)
    >>>
    >>> dataset = datasets.load_digits()
    >>> x_train, x_test, y_train, y_test = train_test_split(
    ...     dataset.data, dataset.target, train_size=0.7
    ... )
    >>>
    >>> nw = PNN(standard_deviation=10, verbose=False)
    >>> nw.train(x_train, y_train)
    >>> result = nw.predict(x_test)
    >>> metrics.accuracy_score(y_test, result)
    0.98888888888888893
    """
    standard_deviation = NonNegativeNumberProperty(default=0.1)

    def __init__(self, **options):
        super(PNN, self).__init__(FAKE_CONNECTION, **options)
        self.output_layer = layers.OutputLayer(1)
        self.classes = None

    def train(self, input_train, target_train):
        LazyLearning.train(self, input_train, target_train)

        if target_train.ndim != 1:
            raise ValueError("Target value must be in 1 dimention")

        classes = self.classes = unique(target_train)
        number_of_classes = classes.size
        row_comb_matrix = self.row_comb_matrix = zeros(
            (number_of_classes, input_train.shape[0])
        )
        class_ratios = self.class_ratios = zeros(number_of_classes)

        for i, class_name in enumerate(classes):
            class_val_positions = (target_train == i)
            row_comb_matrix[i, class_val_positions] = 1
            class_ratios[i] = np_sum(class_val_positions)

    def setup_defaults(self):
        # Remove properties from BaseNetwork
        del self.use_bias
        del self.error
        del self.step
        super(PNN, self).setup_defaults()

    def predict_prob(self, input_data):
        raw_output = self.raw_predict(input_data)
        total_output_sum = raw_output.sum(axis=0).reshape(
            (raw_output.shape[1], 1)
        )
        return raw_output.T / total_output_sum

    def raw_predict(self, input_data):
        super(PNN, self).predict(input_data)

        if self.classes is None:
            raise ValueError("Train network before get prediction")

        input_data_size = input_data.shape[1]
        train_data_size = self.input_train.shape[1]

        if input_data_size != train_data_size:
            raise ValueError("Input data must contains {0} features, got "
                             "{1}".format(train_data_size, input_data_size))

        class_ratios = self.class_ratios
        pdf_outputs = pdf_between_data(self.input_train, input_data,
                                       self.standard_deviation)
        return dot(
            self.row_comb_matrix, pdf_outputs
        ) / class_ratios.reshape((class_ratios.size, 1))

    def predict(self, input_data):
        raw_output = self.raw_predict(input_data)
        return self.classes[raw_output.argmax(axis=0)]



from numpy import reshape, nonzero

from .base import BaseAssociative


__all__ = ('Kohonen',)


class Kohonen(BaseAssociative):
    """Kohonen unsupervised associative Neural Network.
    This algorith similar to :network:`Instar`. Like the instar rule, the
    Kohonen rule allows the weights of a neuron to learn an input vector
    and is therefore suitable for recognition applications. One difference
    that this algorithm is not proportional to output. This Kohonen network
    interpretetion update only weights with non-zero output.

    Notes
    -----
    * Network architecture must contains two layers.

    Parameters
    ----------
    {full_params}

    Methods
    -------
    {unsupervised_train_epochs}
    {full_methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neuralpy import algorithms, layers
    >>>
    >>> np.random.seed(0)
    >>>
    >>> input_data = np.array([
    ...     [0.1961,  0.9806],
    ...     [-0.1961,  0.9806],
    ...     [0.9806,  0.1961],
    ...     [0.9806, -0.1961],
    ...     [-0.5812, -0.8137],
    ...     [-0.8137, -0.5812],
    ... ])
    >>>
    >>> kohonet = algorithms.Kohonen(
    ...     layers.LinearLayer(2) > layers.CompetitiveOutputLayer(3),
    ...     step=0.5,
    ...     verbose=False
    ... )
    >>> kohonet.train(input_data, epochs=100)
    >>> kohonet.predict(input_data)
    array([[ 0.,  1.,  0.],
           [ 0.,  1.,  0.],
           [ 1.,  0.,  0.],
           [ 1.,  0.,  0.],
           [ 0.,  0.,  1.],
           [ 0.,  0.,  1.]])
    """
    def update_indexes(self, layer_output):
        _, index_y = nonzero(layer_output)
        return index_y

    def train_epoch(self, input_train, target_train):
        weight = self.input_layer.weight
        predict = self.predict
        update_indexes = self.update_indexes

        for input_row in input_train:
            input_row = reshape(input_row, (1, input_row.size))
            layer_output = predict(input_row)

            index_y = update_indexes(layer_output)
            self.input_layer.weight[:, index_y] += self.step * (
                input_row.T - weight[:, index_y]
            )



from operator import mul

from numpy import zeros, asmatrix, identity, tile, dot, concatenate

from neuralpy.core.properties import NonNegativeNumberProperty
from neuralpy.functions import mse
from neuralpy.algorithms import Backpropagation


__all__ = ('LevenbergMarquardt',)


class LevenbergMarquardt(Backpropagation):
    """ Levenberg-Marquardt algorithm.

    Parameters
    ----------
    mu : float
        Control invertion for J.T * J matrix, defaults to `0.1`.
    mu_increase_factor : float
        Factor to decrease the mu if update decrese the error, otherwise
        increse mu by the same factor.
    {optimizations}
    {raw_predict_param}
    {full_params}

    Methods
    -------
    {supervised_train}
    {full_methods}

    Examples
    --------
    Simple example

    >>> import numpy as np
    >>> from neuralpy import algorithms
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> lmnet = algorithms.LevenbergMarquardt(
    ...     (2, 3, 1),
    ...     verbose=False
    ... )
    >>> lmnet.train(x_train, y_train)

    Diabets dataset example

    >>> import numpy as np
    >>> from sklearn import datasets, preprocessing
    >>> from sklearn.cross_validation import train_test_split
    >>> from neuralpy import algorithms, layers
    >>> from neuralpy.functions import rmsle
    >>>
    >>> dataset = datasets.load_diabetes()
    >>> data, target = dataset.data, dataset.target
    >>>
    >>> data_scaler = preprocessing.MinMaxScaler()
    >>> target_scaler = preprocessing.MinMaxScaler()
    >>>
    >>> x_train, x_test, y_train, y_test = train_test_split(
    ...     data_scaler.fit_transform(data),
    ...     target_scaler.fit_transform(target),
    ...     train_size=0.85
    ... )
    >>>
    >>> # Network
    ... lmnet = algorithms.LevenbergMarquardt(
    ...     connection=[
    ...         layers.SigmoidLayer(10),
    ...         layers.SigmoidLayer(40),
    ...         layers.OutputLayer(1),
    ...     ],
    ...     mu_increase_factor=2,
    ...     mu=0.1,
    ...     step=0.25,
    ...     show_epoch=10,
    ...     use_bias=False,
    ...     verbose=False
    ... )
    >>> lmnet.train(x_train, y_train, epochs=100)
    >>> y_predict = lmnet.predict(x_test)
    >>>
    >>> error = rmsle(target_scaler.inverse_transform(y_test),
    ...               target_scaler.inverse_transform(y_predict).round())
    >>> error
    0.47548200957888398

    See Also
    --------
    :network:`Backpropagation` : Backpropagation algorithm.
    """
    mu = NonNegativeNumberProperty(default=0.01)
    mu_increase_factor = NonNegativeNumberProperty(default=5, min_size=1)

    def setup_defaults(self):
        super(LevenbergMarquardt, self).setup_defaults()
        # Can use only squared error
        del self.error
        self.error = mse  # `error` isn't a property instance after deletion.

    def init_layers(self):
        super(LevenbergMarquardt, self).init_layers()
        self.n_weights = sum(mul(*layer.size) for layer in self.train_layers)

    def get_jacobian_matrix(self):
        result = []
        for input_data, gradient in zip(self.layer_outputs, self.delta):
            gradient_n_col = gradient.shape[1]
            input_data_n_features = input_data.shape[1]

            repeated_gradients = tile(gradient, (1, input_data_n_features))
            # Total weight matrix size (rows*cols) in layer.
            n_weights = repeated_gradients.shape[1]

            # Matrix `comb` help repeate input data columns for valid
            # combination of input data and gradients for Jacobian matrix.
            # Basicly this matrix and dot product replace 2 inner loops
            # in pure python.
            comb = zeros((input_data_n_features, n_weights))
            for i in range(input_data_n_features):
                comb[i, i * gradient_n_col:(i + 1) * gradient_n_col] = 1

            result.append(repeated_gradients * dot(input_data, comb))

        return concatenate(result, axis=1)

    def update_weights(self, weight_deltas):
        error = asmatrix(self.target_train - self.output_train)
        jacoby = self.get_jacobian_matrix() / error
        jacoby_t = jacoby.T
        smallstep_matrix = self.mu * identity(self.n_weights)

        inverse_hessian = (jacoby_t * jacoby + smallstep_matrix).I
        # Most of all time this order would be much faster than sequantial
        # order without brackets, because number of columns for parameter
        # error would be small.
        jacoby_delta = inverse_hessian * (jacoby_t * error)

        row = 0
        old_weights = self.old_weights = []

        for layer in self.train_layers:
            weight = layer.weight
            weight_in_size, weight_out_size = weight.shape
            old_weights.append(weight.copy())

            for i in range(self.use_bias, weight_in_size):
                weight_delta = jacoby_delta[row:row + weight_out_size, :].T
                weight[i:i + 1, :] -= weight_delta
                row += weight_out_size

            if self.use_bias:
                # Use otside if loop because we setup last column for bias
                # which is at fist row position in weight matrix.
                weight[0:1, :] -= jacoby_delta[row:row + weight_out_size, :].T
                row += weight_out_size

    def after_weight_update(self, input_train, target_train):
        super(LevenbergMarquardt, self).after_weight_update(
            input_train, target_train
        )

        if not self.last_error():
            return

        output_train = self.predict(input_train)
        error = self.error(output_train, target_train)

        if error < self.last_error_in():
            self.mu /= self.mu_increase_factor
            return

        self.mu *= self.mu_increase_factor

        # Error changed in wrong way. Rollback weight updates.
        for i, layer in enumerate(self.train_layers):
            layer.weight = self.old_weights[i]



from numpy import zeros, round as np_round, where

from neuralpy.network.connections import NetworkConnectionError
from neuralpy.layers.base import BaseLayer
from neuralpy.core.properties import (NumberBoundProperty,
                                      NonNegativeIntProperty,
                                      BetweenZeroAndOneProperty)


__all__ = ('OutputLayer', 'CompetitiveOutputLayer', 'StepOutputLayer',
           'RoundOutputLayer')


class OutputLayer(BaseLayer):
    """ Simples output layer class which does not make any transformations.
    Output from this layer is the same as input.

    Parameters
    ----------
    {layer_params}
    """
    def initialize(self, *args, **kwargs):
        return

    def relate_to(self, right_layer):
        raise NetworkConnectionError("Can't connect from output layer")

    def format_output(self, value):
        return value

    def output(self, value):
        return self.format_output(value)


class CompetitiveOutputLayer(OutputLayer):
    """ Competitive layer output. Layer output will return the result where
    all zero values and one value which has greatest value will be one.

    Parameters
    ----------
    {layer_params}
    """
    def format_output(self, value):
        output = zeros(value.shape)
        max_args = value.argmax(axis=1)
        output[range(value.shape[0]), max_args] = 1
        return output


class StepOutputLayer(OutputLayer):
    """ The behaviour for this layer is the same as for step function.

    Parameters
    ----------
    output_bounds : tuple
        Value is must be a tuple which contains two elements where first one
        identify lower output value and the second one - bigger. Defaults
        to ``(0, 1)``.
    critical_point : float
        Critical point is setup step function bias.
    {layer_params}
    """
    output_bounds = NumberBoundProperty(default=(0, 1))
    critical_point = BetweenZeroAndOneProperty(default=0.5)

    def format_output(self, value):
        lower_bound, upper_bound = self.output_bounds
        return where(value < self.critical_point, lower_bound, upper_bound)


class RoundOutputLayer(OutputLayer):
    """ Round output layer value.

    Parameters
    ----------
    decimal_places : int
        The precision in decimal digits for output value.
    {layer_params}
    """
    decimal_places = NonNegativeIntProperty(default=0)

    def format_output(self, value):
        return np_round(value, self.decimal_places)



__all__ = ('SupervisedLearning', 'UnsupervisedLearning', 'LazyLearning')


class SupervisedLearning(object):
    """ Mixin for Supervised Neural Network algorithms.
    """
    def train(self, input_train, target_train, input_test=None,
              target_test=None, epochs=None, epsilon=None):
        self._train(input_train=input_train, target_train=target_train,
                    input_test=input_test, target_test=target_test,
                    epochs=epochs, epsilon=epsilon)


class UnsupervisedLearning(object):
    """ Mixin for Unsupervised Neural Network algorithms.
    """
    def train(self, input_train, epochs=None, epsilon=None):
        self._train(input_train=input_train, target_train=None,
                    input_test=None, target_test=None,
                    epochs=epochs, epsilon=epsilon)

    def predict_prob(self, input_data):
        raise AttributeError("Can't predict probabilities in unsupervised "
                             "network")


class LazyLearning(object):
    """ Mixin for Lazy learning Neural Network algorithms.
    """
    def __init__(self, *args, **kwargs):
        self.input_train = None
        self.target_train = None
        super(LazyLearning, self).__init__(*args, **kwargs)

    def setup_defaults(self):
        del self.shuffle_data

    def train(self, input_train, target_train):
        self.input_train = input_train
        self.target_train = target_train

        if input_train.shape[0] != target_train.shape[0]:
            raise ValueError("Input data size must be the same as "
                             "target data size")

    def train_epoch(self, *args, **kwargs):
        raise AttributeError("This network doesn't have train epochs")

    def predict(self, input_data):
        if self.input_train is None:
            raise ValueError("Train your network before make prediction")



__all__ = ("docs",)


docs = {
    # ------------------------------------ #
    #                Methods               #
    # ------------------------------------ #

    "last_error": """last_error()
        Returns the last error network result after training procedure
        or ``None`` value if you try to get it before network training.
    """,
    "plot_errors": """plot_errors(use_semilog=False)
        Draws the error rate update plot. It always shows network
        learning progress. When you add cross validation data set
        into training function it displays validation data set error as
        separated curve. If parameter ``use_semilog`` is equal to the
        ``True`` value it displays x-axis in logarithmic scale.
    """,
    "predict": """predict(input_data)
        Predict input value.
    """,
    "fit": """fit(\*args, \*\*kwargs)
        The same as ``train`` method.
    """,

    # ------------------------------------ #
    #             Train Methods            #
    # ------------------------------------ #

    "supervised_train": """train(input_train, target_train, input_test=None,\
    target_test=None, epochs=100, epsilon=None):
        Trains network. You can control network training procedure
        iterations with the number of epochs or converge value epsilon.
        Also you can specify ``input_test`` and ``target_test`` and control
        your validation data error on each iteration.
    """,
    "supervised_train_epochs": """train(input_data, target_data, epochs=100):
        Trains network with fixed number of epochs.
    """,
    "unsupervised_train_epochs": """train(input_train, epochs=100):
        Trains network with fixed number of epochs.
    """,
    "unsupervised_train_epsilon": """train(input_train, epsilon=1e-5):
        Trains network until it is converged.
    """,
    "supervised_train_lazy": """train(input_train, target_train):
        Network just stores all the information about the data and use it for \
        the prediction.
    """,

    # ------------------------------------ #
    #              Parameters              #
    # ------------------------------------ #

    "verbose": """verbose : bool
        Verbose output for network. Defaults to ``True``. ``False``
        disables output.
    """,
    "step": """step : float
        Learns step, defaults to ``0.1``.
    """,
    "show_epoch": """show_epoch : int
        This option controls how offen you'd see the epoch updates
        information in the terminal, defaults to ``1``
    """,
    "shuffle_data": """shuffle_data : bool
        If it's ``True`` class shuffles all your training data before
        training your network, defaults to ``True``.
    """,
    "error": """error : function
        Function which controls your training error. defaults to ``mse``
    """,
    "use_bias": """use_bias : bool
        Uses bias in the network, defualts to ``True``.
    """,
    "train_epoch_end_signal": """train_epoch_end_signal : function
        Calls this function when train epoch finishes.
    """,
    "train_end_signal": """train_end_signal : function
        Calls this function when train process finishes.
    """,

    # ------------------------------------ #
    #                 Steps                #
    # ------------------------------------ #
    "first_step": """first_step : float
        Contains initialized step value.
    """,
    "steps": """steps : list of float
        List of steps in the same order as the network layers.
        By default all values are equal to ``step`` parameter.
    """,

    # ------------------------------------ #
    #               Warnings               #
    # ------------------------------------ #
    "bp_depending": """It works with any algorithm based on backpropagation. \
    Class can't work without it.
    """
}


# ------------------------------------ #
#         Complex parameters           #
# ------------------------------------ #


def joindocs(docs, docskeys):
    return ''.join([docs[key] for key in docskeys])


full_params_params = ()
docs.update({
    'full_params': joindocs(
        docs,
        [
            'step', 'show_epoch', 'shuffle_data',
            'error', 'use_bias', 'train_epoch_end_signal',
            'train_end_signal', 'verbose'
        ]
    ),
    'full_signals': joindocs(
        docs, ['train_epoch_end_signal', 'train_end_signal']
    ),
    'full_methods': joindocs(
        docs, ['fit', 'predict', 'last_error', 'plot_errors']
    )
})



from time import time
from itertools import groupby

from numpy import arange
import matplotlib.pyplot as plt

from neuralpy.helpers import preformat_value
from neuralpy.core.base import BaseSkeleton
from neuralpy.core.config import Configurable
from neuralpy.core.properties import (NonNegativeIntProperty, FuncProperty,
                                      NumberProperty, BoolProperty)
from neuralpy.layers import BaseLayer, OutputLayer
from neuralpy.functions import normilize_error_output, mse
from .utils import iter_until_converge, shuffle
from .connections import (FAKE_CONNECTION, LayerConnection,
                          NetworkConnectionError)


__all__ = ('BaseNetwork',)


# ----------------- Signals ---------------- #

def train_epoch_end(network):
    network.logs.data("""
        Epoch {epoch}
        Error in:  {error}
        Error out: {error_out}
        Epoch time: {epoch_time} sec
    """.format(
        epoch=network.epoch,
        error=network.last_error_in() or '-',
        error_out=network.last_error_out() or '-',
        epoch_time=round(network.train_epoch_time, 5)
    ))


def train_end(network):
    network.logs.log("TRAIN", "End train")


def clean_layers(connection):
    """ Clean layers connections and format transform them into one format.
    Also this function validate layers connections.

    Parameters
    ----------
    connection : list, tuple or object
        Layers connetion in different formats.

    Returns
    -------
    object
        Cleaned layers connection.
    """
    if connection == FAKE_CONNECTION:
        return connection

    if isinstance(connection, tuple):
        connection = list(connection)

    islist = isinstance(connection, list)

    if islist and isinstance(connection[0], BaseLayer):
        chain_connection = connection.pop()
        for layer in reversed(connection):
            chain_connection = LayerConnection(layer, chain_connection)
        connection = chain_connection

    elif islist and isinstance(connection[0], LayerConnection):
        pass

    if not isinstance(connection.output_layer, OutputLayer):
        raise NetworkConnectionError("Final layer must be OutputLayer class "
                                     "instance.")

    return connection


class NetworkSignals(Configurable):
    """ Network signals.

    Parameters
    ----------
    {full_signals}
    """
    train_epoch_end_signal = FuncProperty(default=train_epoch_end)
    train_end_signal = FuncProperty(default=train_end)


class BaseNetwork(BaseSkeleton, NetworkSignals):
    """ Base class Network algorithms.

    Parameters
    ----------
    {full_params}

    Methods
    -------
    {plot_errors}
    {last_error}
    """
    error = FuncProperty(default=mse)
    use_bias = BoolProperty(default=True)
    step = NumberProperty(default=0.1)

    # Training settings
    show_epoch = NonNegativeIntProperty(min_size=1, default=1)
    shuffle_data = BoolProperty(default=False)

    def __init__(self, connection, **options):
        self.connection = clean_layers(connection)

        self.errors_in = []
        self.errors_out = []
        self.epoch = 0
        self.train_epoch_time = None

        self.layers = list(self.connection)
        self.input_layer = self.layers[0]
        self.output_layer = self.layers[-1]
        self.train_layers = self.layers[:-1]

        # Setup initialized options
        super(BaseNetwork, self).__init__(**options)
        logs = self.logs

        self.setup_defaults()

        available_classes = [c.__name__ for c in self.__class__.__mro__]

        def classname_grouper(option):
            classname = option[1].class_name
            class_priority = -available_classes.index(classname)
            return (class_priority, classname)

        # Sort and group options by classes
        grouped_options = groupby(
            sorted(self.options.items(), key=classname_grouper),
            key=classname_grouper
        )

        if isinstance(self.connection, LayerConnection):
            logs.header("Network structure")
            logs.log("LAYERS", self.connection)

        # Just display in terminal all network options.
        logs.header("Network options")
        for (_, clsname), class_options in grouped_options:
            if not class_options:
                # When in some class we remove all available attributes
                # we just skip it.
                continue

            logs.simple("{}:".format(clsname))

            for key, data in sorted(class_options):
                if key in options:
                    logger = logs.log
                    value = options[key]
                else:
                    logger = logs.gray_log
                    value = data.value

                logger("OPTION", "{} = {}".format(
                    key, preformat_value(value))
                )
            logs.empty()

        self.init_layers()
        super(BaseNetwork, self).__init__()

    def setup_defaults(self):
        """ Setup default values before populate options.
        """

    # ----------------- Neural Network Layers ---------------- #

    def init_layers(self):
        """ Initialize layers.
        """
        if self.connection == FAKE_CONNECTION:
            return

        for layer in self.layers:
            layer.initialize(with_bias=self.use_bias)

    # ----------------- Neural Network Train ---------------- #

    def _train(self, input_train, target_train=None, input_test=None,
               target_test=None, epochs=None, epsilon=None):

        # ----------- Pre-format target data ----------- #

        if target_train is not None and target_train.ndim == 1:
            target_train = target_train.reshape((target_train.size, 1))

        if target_test is not None and target_test.ndim == 1:
            target_test = target_test.reshape((target_test.size, 1))

        # ----------- Validation ----------- #

        if epochs is None and epsilon is None:
            epochs = 100

        if epochs is not None and epsilon is not None:
            raise ValueError("You can't user `epochs` and `epsilon` "
                             "attributes in one train process.")

        # ----------- Predefine parameters ----------- #

        if epochs is not None:
            self.epoch = 0
            iterepochs = range(self.epoch, epochs)
            last_epoch = epochs - 1
            predict = self.predict
            compute_error_out = (input_test is not None and
                                 target_test is not None)

        if epsilon is not None:
            iterepochs = iter_until_converge(self, epsilon)
            last_epoch = None
            predict = None
            compute_error_out = None

        # ----------- Train process ----------- #

        logs = self.logs
        logs.header("Start train")
        logs.log("TRAIN", "Train data size: {}".format(input_train.shape[0]))
        logs.log("TRAIN", "Number of features: {}".format(
            input_train.shape[1]
        ))

        if epochs is not None:
            logs.log("TRAIN", "Total epochs: {}".format(epochs))

        logs.empty()

        # Optimizations for long loops. Set constant properties to
        # variables.
        errors = self.errors_in
        errors_out = self.errors_out
        show_epoch = self.show_epoch
        shuffle_data = self.shuffle_data

        # Methods
        error_func = self.error
        train_epoch = self.train_epoch
        train_epoch_end_signal = self.train_epoch_end_signal

        for epoch in iterepochs:
            epoch_start_time = time()

            if shuffle_data:
                if target_train is not None:
                    input_train, target_train = shuffle(input_train,
                                                        target_train)
                else:
                    input_train, = shuffle(input_train)

            self.input_train = input_train
            self.target_train = target_train

            try:
                error = train_epoch(input_train, target_train)

                if compute_error_out:
                    error_out = error_func(predict(input_test), target_test)
                    errors_out.append(error_out)

                errors.append(error)
                self.train_epoch_time = time() - epoch_start_time

                if epoch % show_epoch == 0 or epoch == last_epoch:
                    train_epoch_end_signal(self)

                self.epoch = epoch + 1

            except StopIteration as err:
                logs.log("TRAIN", "Epoch #{} stopped. {}".format(self.epoch,
                                                                 str(err)))
                break

        self.train_end_signal(self)

    # ----------------- Errors ----------------- #

    def _last_error(self, errors):
        if errors and errors[-1] is not None:
            return normilize_error_output(errors[-1])

    def last_error_in(self):
        return self._last_error(self.errors_in)

    def last_error(self):
        return self._last_error(self.errors_in)

    def last_error_out(self):
        return self._last_error(self.errors_out)

    def previous_error(self):
        errors = self.errors_in
        return normilize_error_output(errors[-2]) if len(errors) > 2 else None

    def _normalized_errors(self, errors):
        if not len(errors) or isinstance(errors[0], float):
            return errors

        self.logs.warn("Your errors bad formated for plot output. "
                       "They will be normilized.")

        normilized_errors = []
        for error in errors:
            normilized_errors.append(normilize_error_output(error))

        return normilized_errors

    def normalized_errors_in(self):
        return self._normalized_errors(self.errors_in)

    def normalized_errors_out(self):
        return self._normalized_errors(self.errors_out)

    def plot_errors(self, use_semilog=False):
        if not self.errors_in:
            return

        errors_in = self.normalized_errors_in()
        errors_out = self.normalized_errors_out()
        errors_range = arange(len(errors_in))
        plot_function = plt.semilogx if use_semilog else plt.plot

        line_error_in, = plot_function(errors_range, errors_in)

        if errors_out:
            line_error_out, = plot_function(errors_range, errors_out)
            plt.legend(
                [line_error_in, line_error_out],
                ['Error in', 'Error out']
            )

        plt.xlim(0)

        plt.title('Train errors')
        plt.ylabel('Error')
        plt.xlabel('Epoch')

        plt.show()

    # ----------------- Representations ----------------- #

    def get_class_name(self):
        return self.__class__.__name__

    def __repr__(self):
        options = []
        for option_name in self.options:
            option_value = getattr(self, option_name)
            option_value = preformat_value(option_value)

            option_repr = "{}={}".format(option_name, option_value)
            options.append(option_repr)

        classname = self.get_class_name()
        options_repr = ', '.join(options)

        if self.connection != FAKE_CONNECTION:
            return "{}({}, {})".format(classname, self.connection,
                                       options_repr)
        return "{}({})".format(classname, options_repr)



import operator

from numpy import concatenate, reshape, asarray


__all__ = ('matrix_list_in_one_vector', 'vector_to_list_of_matrix')


def matrix_list_in_one_vector(matrix_list):
    """ Function concatenate all matrix from list in a single vector.

    Parameters
    ----------
    matrix_list : list of array-like elements
        List of matrices.

    Returns
    -------
    array-like
        Function will return a single vector wich contains all matrix
        transformed to the vector and concatenated in the same order
        as in the list.

    Examples
    --------
    >>> import numpy as np
    >>> from neuralpy.algorithms.utils import *
    >>>
    >>> a = np.arange(9).reshape((3, 3))
    >>> b = np.array([[10, 0], [0, -1]])
    >>>
    >>> matrix_list_in_one_vector([a, b])
    array([ 0,  1,  2,  3,  4,  5,  6,  7,  8, 10,  0,  0, -1])
    >>> matrix_list_in_one_vector([b, a])
    array([10,  0,  0, -1,  0,  1,  2,  3,  4,  5,  6,  7,  8])
    """
    return concatenate([m.ravel() for m in matrix_list])


def vector_to_list_of_matrix(vector, matrix_sizes):
    """ Funtion split vector into few pices and convert each one to
    a matrix.

    Parameters
    ----------
    vector : arra-like
        The vector which you need to split in few matrices.
    matrix_sizes : list
        List of new matrix shapes.

    Returns
    -------
    list
        List of all matrices built form ``vector`` parameter using
        matrix shapes list.

    Examples
    --------
    >>> import numpy as np
    >>> from neuralpy.algorithms.utils import *
    >>>
    >>> vec = np.array([10, 0, 0, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8])
    >>> vector_to_list_of_matrix(vec, [(2, 2), (3, 3)])
    [array([[10,  0],
           [ 0, -1]]), array([[0, 1, 2],
           [3, 4, 5],
           [6, 7, 8]])]
    """
    start_pos = 0
    matrix = []
    # if we get matrix it be in 2-dimention space,
    # so we need flatten it
    vector = asarray(vector).ravel()

    for size in matrix_sizes:
        vector_length = operator.mul(*size)
        end_pos = start_pos + vector_length

        reshaped_matrix = reshape(vector[start_pos:end_pos], size)
        matrix.append(reshaped_matrix)

        start_pos = end_pos

    return matrix
